from typing import Optional
from typing import overload

from spinta.typing import FileObjectData
from spinta import commands
from spinta.components import Context, Property
from spinta.backends.components import BackendFeatures
from spinta.backends.postgresql.files import DatabaseFile
from spinta.types.datatype import File
from spinta.exceptions import NotFoundError, ItemDoesNotExist
from spinta.backends.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL
from spinta.utils.nestedstruct import flat_dicts_to_nested


@overload
@commands.getone.register(Context, Property, File, PostgreSQL)
def getone(
    context: Context,
    prop: Property,
    dtype: File,
    backend: PostgreSQL,
    *,
    id_: str,
) -> FileObjectData:
    table = backend.get_table(prop.model)
    connection = context.get('transaction').connection
    selectlist = [
        table.c._id,
        table.c._revision,
        table.c[prop.place + '._id'],
        table.c[prop.place + '._content_type'],
        table.c[prop.place + '._size'],
    ]

    if BackendFeatures.FILE_BLOCKS in prop.dtype.backend.features:
        selectlist += [
            table.c[prop.place + '._bsize'],
            table.c[prop.place + '._blocks'],
        ]

    try:
        data = backend.get(connection, selectlist, table.c._id == id_)
    except NotFoundError:
        raise ItemDoesNotExist(dtype, id=id_)

    result = {
        '_type': prop.model_type(),
        '_id': data[table.c._id],
        '_revision': data[table.c._revision],
    }

    data = flat_dicts_to_nested(data)
    result[prop.name] = data[prop.name]
    return commands.cast_backend_to_python(context, prop, backend, result)


@commands.getfile.register()
def getfile(
    context: Context,
    prop: Property,
    dtype: File,
    backend: PostgreSQL,
    *,
    data: FileObjectData,
) -> Optional[bytes]:
    if not data['_blocks']:
        return None

    if len(data['_blocks']) > 1:
        # TODO: Use propper UserError exception.
        raise Exception(
            "File content is to large to retrun it inline. Try accessing "
            "this file directly using subresources API."
        )

    connection = context.get('transaction').connection
    table = backend.get_table(prop, TableType.FILE)
    with DatabaseFile(
        connection,
        table,
        data['_size'],
        data['_blocks'],
        data['_bsize'],
        mode='r',
    ) as f:
        return f.read()



from spinta import commands
from spinta.backends.constants import TableType
from spinta.backends.components import BackendFeatures
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name
from spinta.backends.postgresql.helpers import get_column_name
from spinta.components import Context
from spinta.types.datatype import File
from spinta.migrations import SchemaVersion


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, type(None), File)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: type(None),
    current: File,
):
    if current.backend is backend:
        file_blocks_table_name = get_pg_name(get_table_name(current.prop, TableType.FILE))
        file_table_columns = [
            {
                'name': 'column',
                'args': ['_id', {'name': 'uuid', 'args': []}],
            },
            {
                'name': 'column',
                'args': ['_block', {'name': 'binary', 'args': []}],
            },
        ]
        version.actions.append({
            'type': 'schema',
            'upgrade': {
                'name': 'create_table',
                'args': [file_blocks_table_name] + file_table_columns,
            },
            'downgrade': {
                'name': 'drop_table',
                'args': [file_blocks_table_name],
            },
        })
    pr = [
        {
            'name': 'column',
            'args': [
                f'{get_column_name(current.prop)}._id',
                {
                    'name': 'string',
                    'args': [],
                },
            ]
        },
        {
            'name': 'column',
            'args': [
                f'{get_column_name(current.prop)}._content_type',
                {
                    'name': 'string',
                    'args': [],
                },
            ]
        },
        {
            'name': 'column',
            'args': [
                f'{get_column_name(current.prop)}._size',
                {
                    'name': 'integer',
                    'args': [],
                },
            ]
        }
    ]
    if BackendFeatures.FILE_BLOCKS in current.backend.features:
        pr.extend(
            [
                {
                    'name': 'column',
                    'args': [
                        f'{get_column_name(current.prop)}._bsize',
                        {
                            'name': 'integer',
                            'args': [],
                        },
                    ]
                },
                {
                    'name': 'column',
                    'args': [
                        f'{get_column_name(current.prop)}._blocks',
                        {
                            'name': 'array',
                            'args': [{'name': 'uuid', 'args': []}],
                        },
                    ]
                }
            ]
        )
    return pr



from typing import Union


def asbool(s: Union[str, bool]) -> bool:
    if isinstance(s, bool):
        return s
    if s in ('true', '1', 'on', 'yes'):
        return True
    if s in ('false', '0', 'off', 'no', ''):
        return False
    raise ValueError(f"Expected a boolean value, got {s!r}.")



import pathlib


CONFIG = {
    'config': [],
    'commands': {
        'modules': [
            'spinta.backends',
            'spinta.formats',
            'spinta.commands',
            'spinta.types',
            'spinta.urlparams',
            'spinta.manifests',
            'spinta.datasets',
            'spinta.dimensions',
            'spinta.naming',
            'spinta.ufuncs',
        ],
        'service': {
            'range': 'spinta.commands.helpers:range_',
        },
    },
    'ufuncs': [
        'spinta.ufuncs',
        'spinta.datasets.backends.sql.commands.query',
    ],
    'components': {
        'core': {
            'context': 'spinta.components:Context',
            'config': 'spinta.components:Config',
            'store': 'spinta.components:Store',
        },
        'accesslog': {
            'file': 'spinta.accesslog.file:FileAccessLog',
            'python': 'spinta.accesslog.python:PythonAccessLog',
        },
        'manifests': {
            'backend': 'spinta.manifests.backend.components:BackendManifest',
            'yaml': 'spinta.manifests.yaml.components:YamlManifest',
            'inline': 'spinta.manifests.yaml.components:InlineManifest',
            'tabular': 'spinta.manifests.tabular.components:CsvManifest',
            'csv': 'spinta.manifests.tabular.components:CsvManifest',
            'ascii': 'spinta.manifests.tabular.components:AsciiManifest',
            'xlsx': 'spinta.manifests.tabular.components:XlsxManifest',
            'gsheets': 'spinta.manifests.tabular.components:GsheetsManifest',
        },
        'backends': {
            # In memory backends mostly usable in tests
            'memory': 'spinta.backends.memory.components:Memory',

            # Internal backends
            'postgresql': 'spinta.backends.postgresql.components:PostgreSQL',
            'mongo': 'spinta.backends.mongo.components:Mongo',
            'fs': 'spinta.backends.fs.components:FileSystem',

            # External backends
            # XXX: Probably these should be moved to components.resources?
            'sql': 'spinta.datasets.backends.sql.components:Sql',
            'sqldump': 'spinta.datasets.backends.sqldump.components:SqlDump',
            'csv': 'spinta.datasets.backends.csv.components:Csv',
            'xml': 'spinta.datasets.backends.notimpl.components:BackendNotImplemented',
            'xlsx': 'spinta.datasets.backends.notimpl.components:BackendNotImplemented',
            'json': 'spinta.datasets.backends.notimpl.components:BackendNotImplemented',
            'geojson': 'spinta.datasets.backends.notimpl.components:BackendNotImplemented',
            'html': 'spinta.datasets.backends.notimpl.components:BackendNotImplemented',
        },
        'migrations': {
            'alembic': 'spinta.migrations.schema.alembic:Alembic',
        },
        'nodes': {
            'ns': 'spinta.components:Namespace',
            'model': 'spinta.components:Model',
            'owner': 'spinta.types.owner:Owner',
            'project': 'spinta.types.project:Project',
            'dataset': 'spinta.datasets.components:Dataset',
        },
        'datasets': {
            'resource': 'spinta.datasets.components:Resource',
            'entity': 'spinta.datasets.components:Entity',
            'attribute': 'spinta.datasets.components:Attribute',
        },
        'keymaps': {
            'sqlalchemy': 'spinta.datasets.keymaps.sqlalchemy:SqlAlchemyKeyMap',
        },
        'types': {
            'any': 'spinta.types.datatype:DataType',
            'pk': 'spinta.types.datatype:PrimaryKey',
            'date': 'spinta.types.datatype:Date',
            'time': 'spinta.types.datatype:Time',
            'datetime': 'spinta.types.datatype:DateTime',
            'temporal': 'spinta.types.datatype:DateTime',
            'string': 'spinta.types.datatype:String',
            'binary': 'spinta.types.datatype:Binary',
            'integer': 'spinta.types.datatype:Integer',
            'number': 'spinta.types.datatype:Number',
            'boolean': 'spinta.types.datatype:Boolean',
            'url': 'spinta.types.datatype:URL',
            'uri': 'spinta.types.datatype:URI',
            'image': 'spinta.types.datatype:Image',
            'geometry': 'spinta.types.geometry.components:Geometry',
            'spatial': 'spinta.types.geometry.components:Spatial',
            'ref': 'spinta.types.datatype:Ref',
            'backref': 'spinta.types.datatype:BackRef',
            'generic': 'spinta.types.datatype:Generic',
            'array': 'spinta.types.datatype:Array',
            'object': 'spinta.types.datatype:Object',
            'file': 'spinta.types.datatype:File',
            'rql': 'spinta.types.datatype:RQL',
            'json': 'spinta.types.datatype:JSON',
        },
        'urlparams': {
            'component': 'spinta.urlparams:UrlParams',
        },
        'dimensions': {
            'prefix': 'spinta.dimensions.prefix.components:UriPrefix',
        }
    },
    'exporters': {
        'ascii': 'spinta.formats.ascii.components:Ascii',
        'csv': 'spinta.formats.csv.components:Csv',
        'json': 'spinta.formats.json.components:Json',
        'jsonl': 'spinta.formats.jsonlines.components:JsonLines',
        'html': 'spinta.formats.html.components:Html',
    },
    'accesslog': {
        'type': 'file',
        'file': '/dev/null',
    },
    'manifests': {
        'default': {
            'type': 'memory',
            'backend': '',
            'mode': 'internal',
            'keymap': '',
        },
    },

    'manifest': 'default',

    # Parameters for datasets, for example credentials. Example:
    #
    #   datasets: {default: {gov/vpk: {resource: sql://user:pass@host/db}}
    #
    # There configuraiton parameters should be read directly by datasets.
    'datasets': {},

    # When scanning for manifest YAML files, ignore these files.
    'ignore': [
        '.travis.yml',
        '/prefixes.yml',
        '/schema/',
        '/env/',
    ],

    'debug': False,

    # How much time to wait in seconds for the backends to go up.
    'wait': 30,

    # Configuration path, where clients, keys and other configuration files are
    # stored. Defaults to ~/.config/spinta.
    'config_path': None,

    # Data path, where all data files are stored, like push state, keymap, etc.
    # Defaults to ~/.local/share/spinta.
    'data_path': None,

    # Path to documentation folder. If set will serve the folder at `/docs`.
    'docs_path': None,

    # Credentials file for remote Spinta instances.
    'credentials_file': None,

    'server_url': 'https://example.com/',

    # This prefix will be added to autogenerated scope names.
    'scope_prefix': 'spinta_',

    # Scope name formatter.
    'scope_formatter': 'spinta.auth:get_scope_name',

    # Part of the scope will be replaced with a hash to fit scope_max_length.
    'scope_max_length': 60,

    # If True, then 'id' property is always included into Action.SEARCH result,
    # even if not explicitly asked.
    'always_show_id': False,

    'default_auth_client': 'default',

    # Public JWK key for validating auth bearer tokens.
    'token_validation_key': None,

    # Limit access to specified namespace root.
    'root': None,

    'env': 'prod',

    'environments': {
        'dev': {
            'keymaps.default': {
                'type': 'sqlalchemy',
                'dsn': 'sqlite:///{data_dir}/keymap.db',
            },
            'backends': {
                'default': {
                    'type': 'postgresql',
                    'dsn': 'postgresql://admin:admin123@localhost:54321/spinta',
                    'migrate': 'alembic',
                },
                'mongo': {
                    'type': 'mongo',
                    'dsn': 'mongodb://admin:admin123@localhost:27017/',
                    'db': 'spinta',
                },
                'fs': {
                    'type': 'fs',
                    'path': pathlib.Path() / 'var/files',
                },
            },
            'manifests': {
                # TODO: remove `default` manifest
                'default': {
                    'type': 'yaml',
                    'path': pathlib.Path() / 'tests/manifest',
                    'sync': None,
                    'keymap': 'default',
                    'mode': 'internal',
                    'backend': 'default',
                },
                'yaml': {
                    'type': 'yaml',
                    'backend': 'default',
                    'path': pathlib.Path() / 'tests/manifest',
                },
            },
            'config_path': pathlib.Path('tests/config'),
        },
        'test': {
            'accesslog': {
                'type': 'python',
            },
            'backends': {
                'default': {
                    'type': 'postgresql',
                    'dsn': 'postgresql://admin:admin123@localhost:54321/spinta_tests',
                },
                'mongo': {
                    'type': 'mongo',
                    'dsn': 'mongodb://admin:admin123@localhost:27017/',
                    'db': 'spinta_tests',
                },
                'fs': {
                    'type': 'fs',
                    'path': pathlib.Path() / 'var/files',
                },
            },
            'manifests': {
                # TODO: remove `default` manifest
                'default': {
                    'type': 'yaml',
                    'path': pathlib.Path() / 'tests/manifest',
                    'sync': None,
                    'backend': 'default',
                    'keymap': 'default',
                    'mode': 'internal',
                },
                'yaml': {
                    'type': 'yaml',
                    'backend': 'default',
                    'path': pathlib.Path() / 'tests/manifest',
                },
            },
            'config_path': pathlib.Path('tests/config'),
            'default_auth_client': 'baa448a8-205c-4faa-a048-a10e4b32a136',
        }
    },
}



import importlib
import inspect
from typing import Any
from typing import Type


def importstr(path):
    if ':' not in path:
        raise Exception(
            f"Can't import python path: {path!r}. Python path must be in "
            f"'dotted.path:Name' form."
        )
    module, obj = path.split(':', 1)
    module = importlib.import_module(module)
    obj = getattr(module, obj)
    return obj


def full_class_name(obj: Any) -> str:
    klass: Type
    if not inspect.isclass(obj):
        klass = type(obj)
    else:
        klass = obj
    return f'{klass.__module__}.{klass.__name__}'



from __future__ import annotations

from spinta.components import Model
from spinta.components import Property
from spinta.core.ufuncs import Env


class ChangeModelBase(Env):
    model: Model
    prop: Property




from typing import Optional
from typing import Union
from urllib.parse import urlencode

from geoalchemy2 import WKBElement
from geoalchemy2.shape import to_shape

from pyproj import CRS, Transformer
from shapely.ops import transform

from shapely.geometry import Point

from spinta.types.geometry.constants import WGS84


def get_osm_link(value: WKBElement, srid: Optional[int]) -> Optional[str]:
    if srid is None:
        # If we don't know how to correctly convert to WGS84, then we
        # can't return a link OSM.
        return None

    shape = to_shape(value)

    if srid and srid != WGS84:
        transformer = Transformer.from_crs(
            crs_from=CRS(f'EPSG:{srid}'),
            crs_to=CRS(f'EPSG:{WGS84}'),
            always_xy=True,
        )
        shape = transform(transformer.transform, shape)

    centroid = shape.centroid
    lon, lat = centroid.x, centroid.y
    params = urlencode({'mlat': lat, 'mlon': lon})
    return f'https://www.openstreetmap.org/?{params}#map=19/{lat}/{lon}'


def get_display_value(shape) -> Union[object, str]:
    if isinstance(shape, Point):
        value = shape.wkt
    else:
        # Shorten possibly long WKT values by reducing it to shape type.
        value = shape.type.upper()
    return value



from typing import Any

from spinta.components import Context
from spinta.components import Model
from spinta.core.ufuncs import Bind
from spinta.core.ufuncs import Expr
from spinta.ufuncs.changebase.components import ChangeModelBase
from spinta.ufuncs.components import ForeignProperty


def _re_change_expr(
    env: ChangeModelBase,
    fpr: ForeignProperty,
    expr: Any,
) -> Any:
    if isinstance(expr, Expr):
        args = [_re_change_expr(env, fpr, a) for a in expr.args]
        kwargs = {k: _re_change_expr(env, fpr, v) for k, v in expr.kwargs.items()}
        return type(expr)(expr.name, *args, **kwargs)

    if isinstance(expr, tuple):
        return tuple(_re_change_expr(env, fpr, v) for v in expr)

    if isinstance(expr, list):
        return [_re_change_expr(env, fpr, v) for v in expr]

    if isinstance(expr, dict):
        return {k: _re_change_expr(env, fpr, v) for k, v in expr}

    if isinstance(expr, ForeignProperty):
        return fpr.join(expr).get_bind_expr()

    if isinstance(expr, Bind):
        prop = env.model.properties[expr.name]
        return fpr.push(prop).get_bind_expr()

    return expr


def change_base_model(
    context: Context,
    model: Model,
    fpr: ForeignProperty,
) -> Expr:
    """Rewrite given expr by changing all binds to a new base model.

    """
    expr = model.external.prepare
    env = ChangeModelBase(context)
    env.update(model=model)
    expr = env.resolve(expr)
    return _re_change_expr(env, fpr, expr)



import ujson as json

from typing import Optional

from starlette.requests import Request
from starlette.responses import StreamingResponse

from spinta.components import Context, Action, UrlParams, Model, Node
from spinta import commands
from spinta.formats.jsonlines.components import JsonLines


@commands.render.register(Context, Request, Model, JsonLines)
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: JsonLines,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    return _render(fmt, data, status_code, headers)


@commands.render.register(Context, Request, Node, JsonLines)
def render(
    context: Context,
    request: Request,
    node: Node,
    fmt: JsonLines,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    return _render(fmt, data, status_code, headers)


def _render(fmt: JsonLines, data, status_code: int, headers: dict):
    stream = _stream(data)
    return StreamingResponse(
        stream,
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers
    )


async def _stream(data):
    for row in data:
        yield json.dumps(row, ensure_ascii=False) + '\n'



import pathlib

from spinta.backends.components import BackendFeatures
from spinta.components import DataSubItem, Action
from spinta.types.datatype import File
from spinta.utils.data import take


def prepare_patch_data(
    dtype: File,
    data: DataSubItem,
) -> dict:
    if data.root.action == Action.DELETE:
        patch = {
            '_id': None,
            '_content_type': None,
            '_size': None,
        }
    else:
        patch = take(['_id', '_content_type', '_size'], data.patch)

    if BackendFeatures.FILE_BLOCKS in dtype.backend.features:
        if data.root.action == Action.DELETE:
            patch.update({
                '_blocks': [],
                '_bsize': None,
            })
        else:
            patch.update(take(['_blocks', '_bsize'], data.patch))

    if isinstance(patch.get('_id'), pathlib.Path):
        # On FileSystem backend '_id' is a Path.
        # XXX: It would be nice to decouple this bey visiting each file property
        #      separaterly.
        patch['_id'] = str(patch['_id'])

    return patch



import datetime
import hashlib
import textwrap
import uuid

import itertools
import json
import logging
import pathlib

import pprintpp
import time
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypedDict

import msgpack
import requests
import sqlalchemy as sa
import tqdm
from requests import HTTPError
from typer import Argument
from typer import Context as TyperContext
from typer import Option
from typer import Exit
from typer import echo

from spinta import exceptions
from spinta import spyna
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.data import ModelRow
from spinta.cli.helpers.data import count_rows
from spinta.cli.helpers.data import ensure_data_dir
from spinta.cli.helpers.data import iter_model_rows
from spinta.cli.helpers.store import prepare_manifest
from spinta.client import RemoteClientCredentials
from spinta.client import get_access_token
from spinta.client import get_client_credentials
from spinta.components import Action
from spinta.components import Config
from spinta.components import Context
from spinta.components import Mode
from spinta.components import Model
from spinta.components import Store
from spinta.core.context import configure_context
from spinta.manifests.components import Manifest
from spinta.types.namespace import sort_models_by_refs
from spinta.utils.data import take
from spinta.utils.json import fix_data_for_json
from spinta.utils.nestedstruct import flatten
from spinta.utils.units import tobytes
from spinta.utils.units import toseconds

log = logging.getLogger(__name__)


def push(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Source manifest files to copy from"
    )),
    output: Optional[str] = Option(None, '-o', '--output', help=(
        "Output data to a given location, by default outputs to stdout"
    )),
    credentials: str = Option(None, '--credentials', help=(
        "Credentials file, defaults to {config_path}/credentials.cfg"
    )),
    dataset: str = Option(None, '-d', '--dataset', help=(
        "Push only specified dataset"
    )),
    auth: str = Option(None, '-a', '--auth', help=(
        "Authorize as a client, defaults to {default_auth_client}"
    )),
    limit: int = Option(None, help=(
        "Limit number of rows read from each model"
    )),
    chunk_size: str = Option('1m', help=(
        "Push data in chunks (1b, 1k, 2m, ...), default: 1m"
    )),
    stop_time: str = Option(None, help=(
        "Stop pushing after given time (1s, 1m, 2h, ...), by default does not "
        "stops until all data is pushed"
    )),
    stop_row: int = Option(None, help=(
        "Stop after pushing n rows, by default does not stop until all data "
        "is pushed"
    )),
    state: pathlib.Path = Option(None, help=(
        "Save push state into a file, by default state is saved to "
        "{data_path}/push/{remote}.db SQLite database file"
    )),
    mode: Mode = Option('external', help=(
        "Mode of backend operation, default: external"
    )),
    dry_run: bool = Option(False, '--dry-run', help=(
        "Read data to be pushed, but do not push or write data to the "
        "destination."
    )),
    stop_on_error: bool = Option(False, '--stop-on-error', help=(
        "Exit immediately on first error."
    ))
):
    """Push data to external data store"""
    if chunk_size:
        chunk_size = tobytes(chunk_size)

    if stop_time:
        stop_time = toseconds(stop_time)

    context = configure_context(ctx.obj, manifests, mode=mode)
    store = prepare_manifest(context)
    config: Config = context.get('config')

    if credentials:
        credsfile = pathlib.Path(credentials)
        if not credsfile.exists():
            echo(f"Credentials file {credsfile} does not exit.")
            raise Exit(code=1)
    else:
        credsfile = config.credentials_file
    # TODO: Read client credentials only if a Spinta URL is given.
    creds = get_client_credentials(credsfile, output)

    if not state:
        ensure_data_dir(config.data_path / 'push')
        state = config.data_path / 'push' / f'{creds.remote}.db'

    manifest = store.manifest
    if dataset and dataset not in manifest.datasets:
        echo(str(exceptions.NodeNotFound(manifest, type='dataset', name=dataset)))
        raise Exit(code=1)

    ns = manifest.namespaces['']

    with context:
        client = auth or config.default_auth_client
        require_auth(context, client)

        _attach_backends(context, store, manifest)
        _attach_keymaps(context, store)

        from spinta.types.namespace import traverse_ns_models

        models = traverse_ns_models(context, ns, Action.SEARCH, dataset)
        models = sort_models_by_refs(models)
        models = list(reversed(list(models)))
        counts = count_rows(
            context,
            models,
            limit,
            stop_on_error=stop_on_error,
        )

        if state:
            engine, metadata = _init_push_state(state, models)
            context.attach('push.state.conn', engine.begin)

        rows = iter_model_rows(
            context,
            models,
            counts,
            limit,
            stop_on_error=stop_on_error,
        )
        rows = _prepare_rows_for_push(rows)

        rows = tqdm.tqdm(rows, 'PUSH', ascii=True, total=sum(counts.values()))

        if stop_time:
            rows = _add_stop_time(rows, stop_time)

        if state:
            rows = _check_push_state(context, rows, metadata)

        if stop_row:
            rows = itertools.islice(rows, stop_row)

        rows = _push_to_remote_spinta(rows, creds, chunk_size, dry_run=dry_run)

        if state and not dry_run:
            rows = _save_push_state(context, rows, metadata)

        _push_rows(rows, stop_on_error)


class _PushRow:
    model: Model
    data: Dict[str, Any]
    # SHA1 checksum of data generated with _get_data_rev.
    rev: Optional[str]
    # True if data has already been sent.
    saved: bool = False

    def __init__(self, model: Model, data: Dict[str, Any]):
        self.model = model
        self.data = data
        self.rev = None
        self.saved = False


def _push_rows(rows: Iterator[_PushRow], stop_on_error: bool = False) -> None:
    while True:
        try:
            next(rows)
        except StopIteration:
            break
        except:
            if stop_on_error:
                raise
            log.exception("Error while reading data.")


def _attach_backends(context: Context, store: Store, manifest: Manifest) -> None:
    context.attach('transaction', manifest.backend.transaction)
    backends = set()
    for backend in store.backends.values():
        backends.add(backend.name)
        context.attach(f'transaction.{backend.name}', backend.begin)
    for backend in manifest.backends.values():
        backends.add(backend.name)
        context.attach(f'transaction.{backend.name}', backend.begin)
    for dataset_ in manifest.datasets.values():
        for resource in dataset_.resources.values():
            if resource.backend and resource.backend.name not in backends:
                backends.add(resource.backend.name)
                context.attach(f'transaction.{resource.backend.name}', resource.backend.begin)


def _attach_keymaps(context: Context, store: Store) -> None:
    for keymap in store.keymaps.values():
        context.attach(f'keymap.{keymap.name}', lambda: keymap)


def _prepare_rows_for_push(rows: Iterable[ModelRow]) -> Iterator[_PushRow]:
    for model, row in rows:
        _id = row['_id']
        _type = row['_type']
        where = {
            'name': 'eq',
            'args': [
                {'name': 'bind', 'args': ['_id']},
                _id,
            ]
        }
        payload = {
            '_op': 'upsert',
            '_type': _type,
            '_id': _id,
            '_where': spyna.unparse(where),
            **{k: v for k, v in row.items() if not k.startswith('_')}
        }
        yield _PushRow(model, payload)


def _push_to_remote_spinta(
    rows: Iterable[_PushRow],
    creds: RemoteClientCredentials,
    chunk_size: int,
    *,
    dry_run: bool = False,
    stop_on_error: bool = False,
) -> Iterator[_PushRow]:
    echo(f"Get access token from {creds.server}")
    token = get_access_token(creds)

    session = requests.Session()
    session.headers['Content-Type'] = 'application/json'
    session.headers['Authorization'] = f'Bearer {token}'

    prefix = '{"_data":['
    suffix = ']}'
    slen = len(suffix)
    chunk = prefix
    ready: List[_PushRow] = []

    for row in rows:
        data = fix_data_for_json(row.data)
        data = json.dumps(data, ensure_ascii=False)
        if ready and len(chunk) + len(data) + slen > chunk_size:
            yield from _send_and_receive(
                session,
                creds.server,
                ready,
                chunk + suffix,
                dry_run=dry_run,
                stop_on_error=stop_on_error,
            )
            chunk = prefix
            ready = []
        chunk += (',' if ready else '') + data
        ready.append(row)

    if ready:
        yield from _send_and_receive(
            session,
            creds.server,
            ready,
            chunk + suffix,
            dry_run=dry_run,
            stop_on_error=stop_on_error,
        )


def _send_and_receive(
    session: requests.Session,
    target: str,
    rows: List[_PushRow],
    data: str,
    *,
    dry_run: bool = False,
    stop_on_error: bool = False,
) -> Iterator[_PushRow]:
    if dry_run:
        recv = _send_data_dry_run(data)
    else:
        recv = _send_data(
            session,
            target,
            rows,
            data,
            stop_on_error=stop_on_error,
        )
    yield from _map_sent_and_recv(rows, recv)


def _send_data_dry_run(
    data: str,
) -> Optional[List[Dict[str, Any]]]:
    """Pretend data has been sent to a target location."""
    recv = json.loads(data)['_data']
    for row in recv:
        if '_id' not in row:
            row['_id'] = str(uuid.uuid4())
        row['_rev'] = str(uuid.uuid4())
    return recv


def _send_data(
    session: requests.Session,
    target: str,
    rows: List[_PushRow],
    data: str,
    *,
    stop_on_error: bool = False,
) -> Optional[List[Dict[str, Any]]]:
    data = data.encode('utf-8')

    try:
        resp = session.post(target, data=data)
    except IOError as e:
        log.error(
            (
                "Error when sending and receiving data.%s\n"
                "Error: %s"
            ),
            _get_row_for_error(rows),
            e,
        )
        if stop_on_error:
            raise
        return

    try:
        resp.raise_for_status()
    except HTTPError:
        try:
            recv = resp.json()
        except requests.JSONDecodeError:
            log.error(
                (
                    "Error when sending and receiving data.\n"
                    "Server response (status=%s):\n%s"
                ),
                resp.status_code,
                textwrap.indent(resp.text, '    '),
            )
            if stop_on_error:
                raise
        else:
            errors = recv.get('errors')
            log.error(
                (
                    "Error when sending and receiving data.%s\n"
                    "Server response (status=%s):\n%s"
                ),
                _get_row_for_error(rows, errors),
                resp.status_code,
                textwrap.indent(pprintpp.pformat(recv), '    '),
            )
        if stop_on_error:
            raise
        return

    return resp.json()['_data']


class _ErrorContext(TypedDict):
    # Manifest name
    manifest: str

    # Dataset name
    dataset: str

    # Absolute model name (but does not start with /)
    model: str

    # Property name
    property: str

    # Property data type
    type: str

    # Model name in data source
    entity: str

    # Property name in data source
    attribute: str

    # Full class path of component involved in the error.
    component: str

    # Line number or other location (depends on manifest) node in manifest
    # involved in the error.
    schema: str

    # External object id
    id: str


class _Error(TypedDict):
    # Error type
    type: str

    # Error code (exception class name)
    code: str

    # Error context.
    context: _ErrorContext

    # Message template.
    tempalte: str

    # Message compiled from templates and context.
    message: str


def _get_row_for_error(
    rows: List[_PushRow],
    errors: Optional[List[_Error]] = None,
) -> str:
    message = []
    size = len(rows)

    if errors:
        rows_by_id = {
            row.data['_id']: row
            for row in rows
            if row.data and '_id' in row.data
        }
        for error in errors:
            ctx = error.get('context') or {}
            if 'id' in ctx and ctx['id'] in rows_by_id:
                row = rows_by_id[ctx['id']]
                message += [
                    f" Model {row.model.name}, data:",
                    ' ' + pprintpp.pformat(row.data),
                ]

    if len(message) == 0 and size > 0:
        row = rows[0]
        data = pprintpp.pformat(row.data)
        message += [
            (
                f" Model {row.model.name},"
                f" items in chunk: {size},"
                f" first item in chunk:"
            ),
            data
        ]

    return '\n'.join(message)


def _map_sent_and_recv(
    sent: List[_PushRow],
    recv: Optional[List[Dict[str, Any]]],
) -> Iterator[_PushRow]:
    if recv is None:
        # We don't have a response, because there was an error while
        # communicating with the target server.
        yield from sent
    else:
        assert len(sent) == len(recv), (
            f"len(sent) = {len(sent)}, len(received) = {len(recv)}"
        )
        for sent_row, recv_row in zip(sent, recv):
            assert sent_row.data['_id'] == recv_row['_id'], (
                f"sent._id = {sent_row.data['_id']}, "
                f"received._id = {recv_row['_id']}"
            )
            yield sent_row


def _add_stop_time(rows, stop):
    start = time.time()
    for row in rows:
        yield row
        if time.time() - start > stop:
            break


def _init_push_state(
    file: pathlib.Path,
    models: List[Model],
) -> Tuple[sa.engine.Engine, sa.MetaData]:
    engine = sa.create_engine(f'sqlite:///{file}')
    metadata = sa.MetaData(engine)
    for model in models:
        table = sa.Table(
            model.name, metadata,
            sa.Column('id', sa.Unicode, primary_key=True),
            sa.Column('rev', sa.Unicode),
            sa.Column('pushed', sa.DateTime),
        )
        table.create(checkfirst=True)
    return engine, metadata


def _get_model_type(row: _PushRow) -> str:
    return row.data['_type']


def _get_data_rev(row: _PushRow):
    rev = fix_data_for_json(take(row.data))
    rev = flatten([rev])
    rev = [[k, v] for x in rev for k, v in sorted(x.items())]
    rev = msgpack.dumps(rev, strict_types=True)
    rev = hashlib.sha1(rev).hexdigest()
    return rev


def _check_push_state(
    context: Context,
    rows: Iterable[_PushRow],
    metadata: sa.MetaData,
):
    conn = context.get('push.state.conn')

    for model_type, group in itertools.groupby(rows, key=_get_model_type):
        table = metadata.tables[model_type]

        query = sa.select([table.c.id, table.c.rev])
        saved = {
            state[table.c.id]: state[table.c.rev]
            for state in conn.execute(query)
        }

        for row in group:
            _id = row.data['_id']

            row.rev = _get_data_rev(row)
            row.saved = _id in saved

            if saved.get(_id) == row.rev:
                continue  # Nothing has changed.

            yield row


def _save_push_state(
    context: Context,
    rows: Iterable[_PushRow],
    metadata: sa.MetaData,
) -> Iterator[_PushRow]:
    conn = context.get('push.state.conn')
    for row in rows:
        table = metadata.tables[row.data['_type']]
        if row.saved:
            conn.execute(
                table.update().
                where(table.c.id == row.data['_id']).
                values(
                    id=row.data['_id'],
                    rev=row.rev,
                    pushed=datetime.datetime.now(),
                )
            )
        else:
            conn.execute(
                table.insert().
                values(
                    id=row.data['_id'],
                    rev=row.rev,
                    pushed=datetime.datetime.now(),
                )
            )
        yield row



from __future__ import annotations

from typing import List
from typing import Optional
from typing import Tuple
from typing import overload

import itertools
from typing import Any
from typing import Dict
from typing import TYPE_CHECKING
from typing import cast

from spinta import commands
from spinta import exceptions
from spinta.auth import authorized
from spinta.commands import authorize
from spinta.commands import check
from spinta.commands import load
from spinta.components import Action
from spinta.components import Base
from spinta.components import Context
from spinta.components import Mode
from spinta.components import Model
from spinta.components import Property
from spinta.core.access import link_access_param
from spinta.core.access import load_access_param
from spinta.dimensions.comments.helpers import load_comments
from spinta.dimensions.enum.components import EnumValue
from spinta.dimensions.enum.components import Enums
from spinta.dimensions.enum.helpers import link_enums
from spinta.dimensions.enum.helpers import load_enums
from spinta.dimensions.lang.helpers import load_lang_data
from spinta.exceptions import KeymapNotSet
from spinta.exceptions import UndefinedEnum
from spinta.exceptions import UnknownPropertyType
from spinta.manifests.components import Manifest
from spinta.manifests.tabular.components import PropertyRow
from spinta.nodes import get_node
from spinta.nodes import load_model_properties
from spinta.nodes import load_node
from spinta.types.namespace import load_namespace_from_name
from spinta.units.helpers import is_unit
from spinta.utils.schema import NA

if TYPE_CHECKING:
    from spinta.datasets.components import Attribute


def _load_namespace_from_model(context: Context, manifest: Manifest, model: Model):
    ns = load_namespace_from_name(context, manifest, model.name)
    ns.models[model.model_type()] = model
    model.ns = ns


@load.register(Context, Model, dict, Manifest)
def load(
    context: Context,
    model: Model,
    data: dict,
    manifest: Manifest,
    *,
    source: Manifest = None,
) -> Model:
    model.parent = manifest
    model.manifest = manifest
    model.mode = manifest.mode  # TODO: mode should be inherited from namespace.
    load_node(context, model, data)
    model.lang = load_lang_data(context, model.lang)
    model.comments = load_comments(model, model.comments)

    if model.keymap:
        model.keymap = manifest.store.keymaps[model.keymap]
    else:
        model.keymap = manifest.keymap

    manifest.add_model_endpoint(model)
    _load_namespace_from_model(context, manifest, model)
    load_access_param(model, data.get('access'), itertools.chain(
        [model.ns],
        model.ns.parents(),
    ))
    load_model_properties(context, model, Property, data.get('properties'))

    # XXX: Maybe it is worth to leave possibility to override _id access?
    model.properties['_id'].access = model.access

    config = context.get('config')

    if model.base:
        base = model.base
        model.base = get_node(
            config,
            manifest,
            model.eid,
            base,
            group='bases',
            ctype='model',
            parent=model,
        )
        model.base.parent = model
        load_node(context, model.base, base, parent=model)
        commands.load(context, model.base, base, manifest)

    if model.external:
        external: dict = model.external
        model.external = get_node(
            config,
            manifest,
            model.eid,
            external,
            group='datasets',
            ctype='entity',
            parent=model,
        )
        model.external.model = model
        load_node(context, model.external, external, parent=model)
        commands.load(context, model.external, external, manifest)
    else:
        model.external = None

    return model


@load.register(Context, Base, dict, Manifest)
def load(context: Context, base: Base, data: dict, manifest: Manifest) -> None:
    pass


@overload
@commands.link.register(Context, Model)
def link(context: Context, model: Model):
    # Link external source.
    if model.external:
        commands.link(context, model.external)

        if model.keymap is None:
            raise KeymapNotSet(model)

    # Link model backend.
    if model.backend:
        if model.backend in model.manifest.backends:
            model.backend = model.manifest.backends[model.backend]
        else:
            model.backend = model.manifest.store.backends[model.backend]
    elif (
        model.mode == Mode.external and
        model.external and
        model.external.resource and
        model.external.resource.backend
    ):
        model.backend = model.external.resource.backend
    else:
        model.backend = model.manifest.backend

    if model.external and model.external.dataset:
        link_access_param(model, itertools.chain(
            [model.external.dataset],
            [model.ns],
            model.ns.parents(),
        ))
    else:
        link_access_param(model, itertools.chain(
            [model.ns],
            model.ns.parents(),
        ))

    # Link model properties.
    for prop in model.properties.values():
        commands.link(context, prop)


@overload
@commands.link.register(Context, Base)
def link(context: Context, base: Base):
    base.model = base.parent.manifest.models[base.model]
    base.pk = [
        base.parent.properties[pk]
        for pk in base.pk
    ]


def _parse_dtype_string(dtype: str) -> Tuple[str, List[str]]:
    if '(' in dtype:
        name, args = dtype.split('(', 1)
        args = args.strip().rstrip(')')
        args = [a.strip() for a in args.split(',')]
        return name, args
    else:
        return dtype, []


@load.register(Context, Property, dict, Manifest)
def load(
    context: Context,
    prop: Property,
    data: PropertyRow,
    manifest: Manifest,
) -> Property:
    config = context.get('config')
    prop.type = 'property'
    prop, data = load_node(context, prop, data, mixed=True)
    prop = cast(Property, prop)
    parents = list(itertools.chain(
        [prop.model, prop.model.ns],
        prop.model.ns.parents(),
    ))
    load_access_param(prop, prop.access, parents)
    prop.enums = load_enums(context, [prop] + parents, prop.enums)
    prop.lang = load_lang_data(context, prop.lang)
    prop.comments = load_comments(prop, prop.comments)

    # Parse dtype like geometry(point, 3346)
    if data['type'] is None:
        raise UnknownPropertyType(prop, type=data['type'])
    dtype_type, dtype_args = _parse_dtype_string(data['type'])
    data = {**data, 'type': dtype_type, 'type_args': dtype_args}

    prop.dtype = get_node(
        config,
        manifest,
        prop.model.eid,
        data,
        group='types',
        parent=prop,
    )
    prop.dtype.type = 'type'
    prop.dtype.prop = prop
    load_node(context, prop.dtype, data)
    if prop.model.external:
        prop.external = _load_property_external(context, manifest, prop, prop.external)
    else:
        prop.external = NA
    commands.load(context, prop.dtype, data, manifest)

    unit: Optional[str] = prop.enum
    if unit is None:
        prop.given.enum = None
        prop.given.unit = None
        prop.enum = None
        prop.unit = None
    elif is_unit(prop.dtype, unit):
        prop.given.enum = None
        prop.given.unit = unit
        prop.enum = None
        prop.unit = unit
    else:
        prop.given.enum = unit

    return prop


def _link_prop_enum(
    prop: Property,
) -> Optional[EnumValue]:
    if prop.given.enum:
        enums: List[Enums] = (
            [prop.enums] +
            [prop.model.ns.enums] +
            [ns.enums for ns in prop.model.ns.parents()] +
            [prop.model.manifest.enums]
        )
        for enums_ in enums:
            if enums_ and prop.given.enum in enums_:
                return enums_[prop.given.enum]
        if not is_unit(prop.dtype, prop.given.enum):
            raise UndefinedEnum(prop, name=prop.given.enum)
    elif prop.enums:
        return prop.enums.get('')


@overload
@commands.link.register(Context, Property)
def link(context: Context, prop: Property):
    commands.link(context, prop.dtype)
    if prop.external:
        if isinstance(prop.external, list):
            for external in prop.external:
                commands.link(context, external)
        else:
            commands.link(context, prop.external)

    model = prop.model

    if prop.model.external and prop.model.external.dataset:
        parents = list(itertools.chain(
            [model],
            [model.external.dataset],
            [model.ns],
            model.ns.parents(),
        ))
    else:
        parents = list(itertools.chain(
            [model],
            [model.ns],
            model.ns.parents(),
        ))
    link_access_param(prop, parents)
    link_enums([prop] + parents, prop.enums)
    prop.enum = _link_prop_enum(prop)


def _load_property_external(
    context: Context,
    manifest: Manifest,
    prop: Property,
    data: Any,  # external data
) -> Attribute:
    if not isinstance(data, dict):
        return _load_property_external(context, manifest, prop, {'name': data})

    config = context.get('config')
    external: Attribute = get_node(
        config,
        manifest,
        prop.model.eid,
        data,
        group='datasets',
        ctype='attribute',
        parent=prop,
    )
    load_node(context, external, data, parent=prop)
    commands.load(context, external, data, manifest)
    return external


@load.register(Context, Model, dict)
def load(context: Context, model: Model, data: dict) -> dict:
    # check that given data does not have more keys, than model's schema
    non_hidden_keys = []
    for key, prop in model.properties.items():
        if not prop.hidden:
            non_hidden_keys.append(key)

    unknown_props = set(data.keys()) - set(non_hidden_keys)
    if unknown_props:
        raise exceptions.MultipleErrors(
            exceptions.FieldNotInResource(model, property=prop)
            for prop in sorted(unknown_props)
        )

    result = {}
    for name, prop in model.properties.items():
        value = data.get(name, NA)
        value = load(context, prop.dtype, value)
        if value is not NA:
            result[name] = value
    return result


@load.register(Context, Property, object)
def load(context: Context, prop: Property, value: object) -> object:
    value = _prepare_prop_data(prop.name, value)
    value[prop.name] = load(context, prop.dtype, value[prop.name])
    return value


def _prepare_prop_data(name: str, data: dict):
    return {
        **{
            k: v
            for k, v in data.items()
            if k.startswith('_') and k not in ('_id', '_content_type')
        },
        name: {
            k: v
            for k, v in data.items()
            if not k.startswith('_') or k in ('_id', '_content_type')
        }
    }


@check.register(Context, Model)
def check(context: Context, model: Model):
    if '_id' not in model.properties:
        raise exceptions.MissingRequiredProperty(model, prop='_id')

    for prop in model.properties.values():
        commands.check(context, prop)


@check.register(Context, Property)
def check(context: Context, prop: Property):
    if prop.enum:
        for value, item in prop.enum.items():
            commands.check(context, item, prop.dtype, item.prepare)


@authorize.register(Context, Action, Model)
def authorize(context: Context, action: Action, model: Model):
    authorized(context, model, action, throw=True)


@authorize.register(Context, Action, Property)
def authorize(context: Context, action: Action, prop: Property):
    authorized(context, prop, action, throw=True)


@overload
@commands.get_error_context.register(Model)
def get_error_context(model: Model, *, prefix='this') -> Dict[str, str]:
    context = commands.get_error_context(model.manifest, prefix=f'{prefix}.manifest')
    context['schema'] = f'{prefix}.get_eid_for_error_context()'
    context['model'] = f'{prefix}.name'
    context['dataset'] = f'{prefix}.external.dataset.name'
    context['resource'] = f'{prefix}.external.resource.name'
    context['resource.backend'] = f'{prefix}.external.resource.backend.name'
    context['entity'] = f'{prefix}.external.name'
    return context


@overload
@commands.get_error_context.register(Property)
def get_error_context(prop: Property, *, prefix='this') -> Dict[str, str]:
    context = commands.get_error_context(prop.model, prefix=f'{prefix}.model')
    context['property'] = f'{prefix}.place'
    context['attribute'] = f'{prefix}.external.name'
    return context



import uuid
import urllib.parse
from typing import Dict, Any
from typing import Iterator

import dask.dataframe as dd
from dask.dataframe import DataFrame

from spinta import commands
from spinta.core.ufuncs import Expr
from spinta.typing import ObjectData
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.components import Model
from spinta.components import Property
from spinta.datasets.utils import iterparams
from spinta.datasets.backends.csv.components import Csv
from spinta.datasets.keymaps.components import KeyMap
from spinta.types.datatype import PrimaryKey
from spinta.types.datatype import Ref
from spinta.datasets.backends.sql.commands.query import Selected
from spinta.dimensions.enum.helpers import get_prop_enum
from spinta.exceptions import ValueNotInEnum
from spinta.utils.schema import NA
from spinta.utils.data import take
from spinta.auth import authorized
from spinta.components import Action
from spinta.exceptions import NoExternalName
from spinta.exceptions import PropertyNotFound


@commands.load.register(Context, Csv, dict)
def load(context: Context, backend: Csv, config: Dict[str, Any]):
    pass


@commands.prepare.register(Context, Csv, Manifest)
def prepare(context: Context, backend: Csv, manifest: Manifest):
    pass


@commands.bootstrap.register(Context, Csv)
def bootstrap(context: Context, backend: Csv):
    pass


def _resolve_expr(
    context: Context,
    row: Dict[str, Any],
    sel: Selected,
) -> Any:
    if sel.item is None:
        val = None
    else:
        val = row[sel.item]
    raise NotImplementedError
    return val
    # env = SqlResultBuilder(context).init(val, sel.prop, row)
    # return env.resolve(sel.prep)


def _get_row_value(context: Context, row: Dict[str, Any], sel: Any) -> Any:
    if isinstance(sel, Selected):
        if isinstance(sel.prep, Expr):
            val = _resolve_expr(context, row, sel)
        elif sel.prep is not NA:
            val = _get_row_value(context, row, sel.prep)
        else:
            val = row[sel.item]

        if enum := get_prop_enum(sel.prop):
            if val is None:
                pass
            elif str(val) in enum:
                item = enum[str(val)]
                if item.prepare is not NA:
                    val = item.prepare
            else:
                raise ValueNotInEnum(sel.prop, value=val)

        return val
    if isinstance(sel, tuple):
        return tuple(_get_row_value(context, row, v) for v in sel)
    if isinstance(sel, list):
        return [_get_row_value(context, row, v) for v in sel]
    if isinstance(sel, dict):
        return {k: _get_row_value(context, row, v) for k, v in sel.items()}
    return sel


def _select(prop: Property, df: DataFrame):
    if prop.external is None or not prop.external.name:
        raise NoExternalName(prop)
    column = prop.external.name
    if column not in df.columns:
        raise PropertyNotFound(
            prop.model,
            property=prop.name,
            external=prop.external.name,
        )
    return Selected(
        item=column,
        prop=prop,
    )


def _select_primary_key(model: Model, df: DataFrame):
    pkeys = model.external.pkeys

    if not pkeys:
        # If primary key is not specified use all properties to uniquely
        # identify row.
        pkeys = take(model.properties).values()

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = _select(prop, df)
    else:
        result = [
            _select(prop, df)
            for prop in pkeys
        ]
    return Selected(prop=model.properties['_id'], prep=result)


def _get_selected(
    context: Context,
    model: Model,
    df: DataFrame,
) -> Dict[str, Selected]:
    selected: Dict[str, Selected] = {}
    for prop in take(['_id', all], model.properties).values():
        if authorized(context, prop, Action.GETALL):
            if isinstance(prop.dtype, PrimaryKey):
                sel = _select_primary_key(model, df)
            else:
                sel = _select(prop, df)
            selected[prop.place] = sel
    return selected


@commands.getall.register(Context, Model, Csv)
def getall(
    context: Context,
    model: Model,
    backend: Csv,
    *,
    query: Expr = None,
) -> Iterator[ObjectData]:
    base = model.external.resource.external
    keymap: KeyMap = context.get(f'keymap.{model.keymap.name}')
    for params in iterparams(context, model):
        if base:
            url = urllib.parse.urljoin(base, model.external.name)
        else:
            url = model.external.name
        url = url.format(**params)
        df = dd.read_csv(url)
        selected: Dict[str, Selected] = _get_selected(context, model, df)
        for row in df.itertuples(index=False):
            row = row._asdict()
            res = {
                '_type': model.model_type(),
            }
            for key, sel in selected.items():
                val = _get_row_value(context, row, sel)
                if sel.prop:
                    if isinstance(sel.prop.dtype, PrimaryKey):
                        val = keymap.encode(sel.prop.model.model_type(), val)
                    elif isinstance(sel.prop.dtype, Ref):
                        val = keymap.encode(sel.prop.dtype.model.model_type(), val)
                        val = {'_id': val}
                res[key] = val
            res = commands.cast_backend_to_python(context, model, backend, res)
            yield res


@commands.wait.register(Context, Csv)
def wait(context: Context, backend: Csv, *, fail: bool = False) -> bool:
    return True



from typing import Iterator

from spinta import commands
from spinta.core.ufuncs import Expr
from spinta.components import Context
from spinta.components import Model
from spinta.typing import ObjectData
from spinta.backends.memory.components import Memory
from spinta.backends.constants import TableType
from spinta.backends.helpers import get_table_name


@commands.getall.register(Context, Model, Memory)
def getall(
    context: Context,
    model: Model,
    db: Memory,
    *,
    query: Expr = None,
) -> Iterator[ObjectData]:
    table = get_table_name(model)
    return db.data[table].values()


@commands.getone.register(Context, Model, Memory)
def getone(
    context: Context,
    model: Model,
    db: Memory,
    *,
    id_: str,
) -> ObjectData:
    table = get_table_name(model)
    return db.data[table][id_]


@commands.changes.register(Context, Model, Memory)
def changes(
    context: Context,
    model: Model,
    db: Memory,
    *,
    id_: str = None,
    limit: int = 100,
    offset: int = -10,
):
    table = get_table_name(model, TableType.CHANGELOG)
    return db.data[table].values()



from typing import TypedDict
from typing import Optional
from typing import List


class ObjectData(TypedDict):
    _type: str
    _id: str
    _revision: str


class FileObjectData(ObjectData):
    _content: Optional[bytes]           # File content
    _content_type: Optional[str]        # File mime content type
    _size: Optional[int]                # File size in bytes

    # Following properties are internal metadata, should not be exposed.
    _bsize: Optional[int]               # Single block size in bytes
    _blocks: Optional[List[str]]        # List of block ids



import pathlib
from io import StringIO
from typing import Tuple
from typing import Union

from pytest import FixtureRequest

from spinta import commands
from spinta.cli.helpers.store import load_store
from spinta.components import Context
from spinta.components import Store
from spinta.core.config import RawConfig
from spinta.core.config import configure_rc
from spinta.manifests.components import Manifest
from spinta.manifests.components import ManifestPath
from spinta.manifests.tabular.helpers import normalizes_columns
from spinta.manifests.tabular.helpers import render_tabular_manifest
from spinta.manifests.tabular.helpers import striptable
from spinta.testing.context import TestContext
from spinta.testing.context import create_test_context


def compare_manifest(manifest: Manifest, expected: str) -> Tuple[str, str]:
    expected = striptable(expected)
    if expected:
        header = expected.splitlines()[0]
        cols = normalizes_columns(header.split('|'))
        sizes = {c: len(x) - 2 for c, x in zip(cols, f' {header} '.split('|'))}
        actual = render_tabular_manifest(manifest, cols, sizes=sizes)
    else:
        actual = render_tabular_manifest(manifest)
    return actual, expected


def load_manifest_get_context(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str] = None,
    *,
    load_internal: bool = False,
    request: FixtureRequest = None,
    **kwargs,
) -> TestContext:
    if isinstance(manifest, pathlib.Path):
        manifests = [str(manifest)]
    elif isinstance(manifest, str) and '|' in manifest:
        manifests = [ManifestPath(
            type='ascii',
            file=StringIO(manifest),
        )]
    elif manifest:
        manifests = [manifest]
    else:
        manifests = manifest

    rc = configure_rc(rc, manifests, **kwargs)
    context = create_test_context(rc, request)
    store = load_store(context, verbose=False, ensure_config_dir=False)
    commands.load(context, store.manifest, load_internal=load_internal)
    commands.link(context, store.manifest)
    return context


def load_manifest(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str],
    **kwargs,
) -> Manifest:
    context = load_manifest_get_context(rc, manifest, **kwargs)
    store: Store = context.get('store')
    return store.manifest


def load_manifest_and_context(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str] = None,
    **kwargs,
) -> Tuple[Context, Manifest]:
    context = load_manifest_get_context(rc, manifest, **kwargs)
    store: Store = context.get('store')
    return context, store.manifest


def prepare_manifest(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str],
    *,
    request: FixtureRequest = None,
    load_internal: bool = True,
    **kwargs,
) -> Tuple[TestContext, Manifest]:
    context = load_manifest_get_context(
        rc, manifest,
        request=request,
        load_internal=load_internal,
        **kwargs,
    )
    store: Store = context.get('store')
    commands.check(context, store.manifest)
    commands.prepare(context, store.manifest)
    return context, store.manifest


def bootstrap_manifest(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str],
    *,
    request: FixtureRequest = None,
    load_internal: bool = True,
    **kwargs,
) -> TestContext:
    context = load_manifest_get_context(
        rc, manifest,
        request=request,
        load_internal=load_internal,
        **kwargs,
    )
    store: Store = context.get('store')
    commands.check(context, store.manifest)
    commands.prepare(context, store.manifest)
    commands.bootstrap(context, store.manifest)
    context.loaded = True
    return context



import sqlalchemy as sa

from sqlalchemy.dialects.postgresql import JSONB

from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Array
from spinta.backends.constants import TableType
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name


@commands.prepare.register(Context, PostgreSQL, Array)
def prepare(context: Context, backend: PostgreSQL, dtype: Array):
    prop = dtype.prop

    columns = commands.prepare(context, backend, dtype.items)
    assert columns is not None
    if not isinstance(columns, list):
        columns = [columns]

    pkey_type = commands.get_primary_key_type(context, backend)

    # TODO: When all list items will have unique id, also add reference to
    #       parent list id.
    # if prop.list:
    #     parent_list_table_name = get_pg_name(
    #         get_table_name(prop.list, TableType.LIST)
    #     )
    #     columns += [
    #         # Parent list table id.
    #         sa.Column('_lid', pkey_type, sa.ForeignKey(
    #             f'{parent_list_table_name}._id', ondelete='CASCADE',
    #         )),
    #     ]

    name = get_pg_name(get_table_name(prop, TableType.LIST))
    main_table_name = get_pg_name(get_table_name(prop.model))
    table = sa.Table(
        name, backend.schema,
        # TODO: List tables eventually will have _id in order to uniquelly
        #       identify list item.
        # sa.Column('_id', pkey_type, primary_key=True),
        sa.Column('_txn', pkey_type, index=True),
        # Main table id (resource id).
        sa.Column('_rid', pkey_type, sa.ForeignKey(
            f'{main_table_name}._id', ondelete='CASCADE',
        )),
        *columns,
    )
    backend.add_table(table, prop, TableType.LIST)

    if prop.list is None:
        # For fast whole resource access we also store whole list in a JSONB.
        return sa.Column(prop.place, JSONB)



from spinta import commands
from spinta.components import Context
from spinta.backends.memory.components import Memory


@commands.wait.register(Context, Memory)
def wait(context: Context, backend: Memory, *, fail: bool = False):
    return True



from spinta.components import Node


class Dimension(Node):
    pass


class UriPrefix(Dimension):
    type: str = 'prefix'
    id: str = None
    eid: str = None
    name: str = None
    uri: str = None
    title: str = None
    description: str = None

    schema = {
        'id': {'type': 'string'},
        'eid': {'type': 'string'},
        'type': {'type': 'string'},
        'name': {'type': 'string'},
        'uri': {'type': 'string'},
        'title': {'type': 'string'},
        'description': {'type': 'string'},
    }

    def __repr__(self):
        return f"<{type(self).__name__} name={self.name!r}>"



from spinta.formats.components import Format


class IterableFile:

    def __init__(self):
        self.writes = []

    def __iter__(self):
        yield from self.writes
        self.writes = []

    def write(self, data):
        self.writes.append(data)


class Csv(Format):
    content_type = 'text/csv'
    accept_types = {
        'text/csv',
    }
    params = {}



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import Set
from typing import TYPE_CHECKING, List, Optional, AsyncIterator, Union

import enum
import contextlib
import dataclasses
import pathlib
from typing import Type
from typing import TypedDict

from spinta import exceptions
from spinta.dimensions.lang.components import LangData
from spinta.units.components import Unit
from spinta.utils.schema import NA
from spinta.core.enums import Access
from spinta.datasets.enums import Level

if TYPE_CHECKING:
    from spinta.backends.components import Backend
    from spinta.types.datatype import DataType
    from spinta.manifests.components import Manifest
    from spinta.manifests.internal.components import InternalManifest
    from spinta.datasets.components import Attribute
    from spinta.datasets.components import Entity
    from spinta.datasets.keymaps.components import KeyMap
    from spinta.dimensions.enum.components import Enums
    from spinta.dimensions.enum.components import EnumValue
    from spinta.core.config import RawConfig
    from spinta.accesslog import AccessLog
    from spinta.formats.components import Format
    from spinta.dimensions.comments.components import Comment


class Context:
    _name: str
    _parent: Optional[Context]
    _exitstack: List[Optional[contextlib.ExitStack]]
    _local_names: List[Set[str]]

    def __init__(self, name: str, parent: Context = None):
        self._name = name
        self._parent = parent

        # Bellow all attributes are lists, these lists are context state stacks.
        # When context is forked or new state is activated, then new item is
        # appended to each list, possibly taking shallow copy from previous
        # state. This way, each state can be changed independently.

        # Names defined in current context. Names from inherited context are not
        # listed here.
        self._local_names = [set()]
        self._exitstack = [None]

        if parent:
            # cmgrs are copied in __enter__() method.
            self._cmgrs = [{}]
            self._factory = [parent._factory[-1].copy()]
            self._names = [parent._names[-1].copy()]

            # We only copy explicitly set keys, and exclude keys coming from
            # `bind` or `attach`.
            copy_keys = set(parent._context[-1]) - (
                set(parent._cmgrs[-1]) |
                set(parent._factory[-1])
            )
            self._context = [{k: parent._context[-1][k] for k in copy_keys}]
        else:
            self._cmgrs = [{}]
            self._factory = [{}]
            self._context = [{}]
            self._names = [set()]

    def __repr__(self):
        name = []
        parent = self
        while parent is not None:
            name.append(f'{parent._name}:{len(parent._context) - 1}')
            parent = parent._parent
        name = ' < '.join(reversed(name))
        return (
            f'<{self.__class__.__module__}.{self.__class__.__name__}({name}) '
            f'at 0x{id(self):02x}>'
        )

    def __enter__(self):
        self._context.append(self._context[-1].copy())
        self._exitstack.append(contextlib.ExitStack())
        self._names.append(self._names[-1].copy())
        self._local_names.append(set())
        self._factory.append({})
        if self._parent and len(self._cmgrs) == 1:
            # We delay copying cmgrs from parent, because ExitStack is only
            # available from inside with block.
            self._cmgrs.append(self._parent._cmgrs[-1].copy())
        else:
            self._cmgrs.append({})
        return self

    def __exit__(self, *exc):
        self._context.pop()
        self._exitstack.pop().close()
        self._names.pop()
        self._local_names.pop()
        self._factory.pop()
        self._cmgrs.pop()

    def fork(self, name) -> Context:
        """Fork this context manager by creating new state.

        This will create new Context class instance based on current state.

        `name` argument is used to identify forked state, this is mostly used
        for debugging, because usually you want to know which state you are
        currently in.

        Forking can be used, when you need to pass context to a function, that
        is executed concurrently.

        Example:

            def hard_answer():
                return 42

            base = Context('base')
            base.set('easy_answer', 42)
            base.bind('hard_answer', hard_answer)

            with base.fork('fork') as fork:

                # Takes answer from cache, because it was set.
                fork.get('easy_answer')

                # Calls `hard_answer` function, even if it was already called
                # for `base`, because we can't modify `base` to ensure thread
                # safety.
                fork.get('hard_answer')

            # Calls `hard_answer` function, even it was already called in
            # `fork`, to ensure thread safety.
            base.get('hard_answer')

            # Takes hard answer from cache.
            base.get('hard_answer')

        """
        return type(self)(name, self)

    def bind(self, name, factory, *args, **kwargs):
        """Bind a callable to the context.

        Value returned by this callable will be cached and callable will be
        called only when first accessed. Subsequent access to this context name
        will return value from cache.
        """
        self._set_local_name(name)
        self._factory[-1][name] = (factory, args, kwargs)

    def attach(self, name, factory, *args, **kwargs):
        """Attach new context manager factory to this state.

        Attached context manager factory is lazy, it will be created only when
        accessed for the first time. Active context managers will exit, when
        whole context exits.

        More information about context managers:

        https://www.python.org/dev/peps/pep-0343/
        https://docs.python.org/3/library/contextlib.html
        """
        assert self._exitstack[-1] is not None, (
            "You can attach only inside `with context:` block."
        )
        self._set_local_name(name)
        self._cmgrs[-1][name] = (factory, args, kwargs)

    def set(self, name, value):
        """Set `name` to `value` in current context."""
        self._set_local_name(name)
        self._context[-1][name] = value
        return value

    def get(self, name):
        """Get a value from context.

        If value was previously set, just return it.

        If value was bound as a callback, then call the callback once and set
        returned value to the context.

        If value was attached as context manager, then find where it was
        attached, enter attached context and return its value.

        Raises error, if given `name` was not set, bound or attached previously.

        Example with bind:

            # State #1
            context = Context('base')

            # Bind a factory on state #1
            context.bind('foo', bar)

            with context:
                # State #2

                with context:
                    # State #3

                    # After calling `bar` function `foo` is cached in current
                    # state #3 and in previous state #1.
                    context.get('foo')

                # Back to state #2. State #3 is removed along with cached `foo`
                # value.

                # Takes 'foo' value from state #1 cache, `bar` function will not
                # be called.
                context.get('foo')

        """

        if name in self._context[-1]:
            return self._context[-1][name]

        # If value is not in current state, then we need to find a factory or
        # a context manager (cmgr) and get value from there. When `name` is
        # found as factory or cmgr, then first we set value in previous state
        # and then set same value in current state.
        stacks = [
            (self._factory, self._get_factory_value),
            (self._cmgrs, self._get_cmgr_value),
        ]
        for stack, value_getter in stacks:
            for state in range(len(stack) - 1, -1, -1):
                if name in stack[state]:
                    if name not in self._context[state]:
                        # Get value and set it on a previous state. This way, if
                        # we exit from current state, value stays in previous
                        # state an can be reused by others state between
                        # previous and current state.
                        self._context[state][name] = value_getter(state, name)
                    if len(self._context) - 1 > state:
                        # If value was not found in current state, then update
                        # current state with value from previous state.
                        self._context[-1][name] = self._context[state][name]
                    return self._context[-1][name]

        raise Exception(f"Unknown context variable {name!r}.")

    def _get_factory_value(self, state, name):
        factory, args, kwargs = self._factory[state][name]
        return factory(*args, **kwargs)

    def _get_cmgr_value(self, state, name):
        factory, args, kwargs = self._cmgrs[state][name]
        cmgr = factory(*args, **kwargs)
        return self._exitstack[state].enter_context(cmgr)

    def has(self, name, local=False, value=False):
        """Check if given name exists in context.

        If `local` is `True`, check only names defined in this context, do not
        look in values defined in parent contexts.

        If `value` is `True`, check only evaluated names. Exclude all bound or
        attached values, that has not yet been accessed.
        """
        if local and value:
            return name in self._local_names[-1] and name in self._context[-1]
        if local:
            return name in self._local_names[-1]
        if value:
            return name in self._context[-1]
        return name in self._names[-1]

    def _set_local_name(self, name):
        # Prevent redefining local names, but allow to redefine inherited names.
        if name in self._local_names[-1]:
            raise Exception(f"Context variable {name!r} has been already set.")
        self._local_names[-1].add(name)
        self._names[-1].add(name)


class _CommandsConfig:

    def __init__(self):
        self.modules = []
        self.pull = {}


class Store:
    """Data store

    Contains all essential objects like manifest, backends, access log, etc.

    """
    manifest: Manifest = None
    internal: InternalManifest = None
    accesslog: AccessLog = None
    backends: Dict[str, Backend]

    def __init__(self):
        self.config = None
        self.keymaps = {}
        self.backends = {}


class Component:
    schema = {}


class Node(Component):
    schema = {}

    type: str = None
    name: str = None
    parent: Node = None
    manifest: Manifest = None
    path: pathlib.Path = None

    def __repr__(self) -> str:
        return f'<{self.__class__.__module__}.{self.__class__.__name__}(name={self.name!r})>'

    def __hash__(self) -> int:
        # This is a recursive hash, goes down to all parents. There can be nodes
        # with same type and name on different parts of nodes tree, that is why
        # we have to also include parents.
        return hash((self.type, self.name, self.parent))

    def node_type(self):
        """Return node type.

        Usually node type is same as `Node.type` attribute, but in some cases,
        there can be same `Node.type`, but with a different behaviour. For
        example there are two types of `model` nodes, one is
        `spinta.types.dataset.Model` and other is `spinta.components.Model`.
        Both these nodes have `model` type, but `Node.node_type()` returns
        different types, `model:dataset` and `model`.

        In other words, there can be several types of model nodes, they mostly
        act like normal models, but they are implemented differently.
        """
        return self.type

    def model_type(self):
        """Return model name and specifier.

        This is a full and unique model type, used to identify a specific model.
        """
        specifier = self.model_specifier()
        if specifier:
            return f'{self.name}/{specifier}'
        else:
            return self.name

    def model_specifier(self):
        """Return model specifier.

        There can by sever different kinds of models. For example
        `model:dataset` always has a specifier, that looks like this
        `:dataset/dsname`, also, `ns` models have `:ns` specifier.
        """
        return ''

    @property
    def basename(self):
        return self.name.split('/')[-1]


# MetaData entry ID can be file path, uuid, table row id of a Model, Dataset,
# etc, depends on manifest type.
EntryId = Union[int, str, pathlib.Path]


class MetaData(Node):
    """Manifest metadata entry.

    This is a top level manifest node like Model, Dataset, Project, Owner.
    """

    eid: EntryId

    schema = {
        'type': {'type': 'string', 'required': True},
        'name': {'type': 'string', 'required': True},
        'id': {'type': 'string'},
        # FIXME: `eid` should be here, but currently it overrides eid, coming
        #        from somewhere else, this has to be fixed.
        # 'eid': {'type': 'string'},
        'version': {'type': 'string'},
        'title': {'type': 'string'},
        'description': {},
        'lang': {'type': 'object'},
    }

    def get_eid_for_error_context(self):
        if (
            isinstance(self.eid, pathlib.Path) and
            self.manifest and
            isinstance(getattr(self.manifest, 'path', None), pathlib.Path)
        ):
            # XXX: Didn't wanted to create command just for this so added na if.
            return str(self.eid.relative_to(self.manifest.path))
        else:
            return str(self.eid)


class NamespaceGiven:
    access: str = None


class Namespace(MetaData):
    access: Access
    keymap: KeyMap = None
    names: Dict[str, Namespace]
    models: Dict[str, Model]
    backend: Backend = None
    parent: Union[Namespace, Manifest]
    title: str
    description: str
    # Namespaces generated from model name.
    generated: bool = False
    given: NamespaceGiven
    lang: LangData = None
    enums: Enums = None

    def __init__(self):
        self.given = NamespaceGiven()

    def model_specifier(self):
        return ':ns'

    def parents(self) -> Iterator[Namespace]:
        ns = self.parent
        while isinstance(ns, Namespace):
            yield ns
            ns = ns.parent

    def is_root(self) -> bool:
        # TODO: Move Namespace component to spinta.namespaces
        from spinta.manifests.components import Manifest
        return isinstance(self.parent, Manifest)


class Base(Node):
    model: Model        # a.base.b - here `model` is `b`
    parent: Model       # a.base.b - here `parent` is `a`
    pk: List[Property]  # a.base.b - list of properties of `a` model
    lang: LangData = None

    schema = {
        'model': {'type': 'string'},
        'pk': {'type': 'string'},
        'lang': {'type': 'object'},
    }


class ModelGiven:
    access: str = None


class Model(MetaData):
    id: str
    level: Level
    access: Access
    title: str
    description: str
    ns: Namespace
    endpoint: str = None
    external: Entity = None
    properties: Dict[str, Property]
    mode: Mode = None
    given: ModelGiven
    lang: LangData = None
    comments: List[Comment] = None

    schema = {
        'keymap': {'type': 'string'},
        'backend': {'type': 'string'},
        'unique': {'default': []},
        'base': {},
        'link': {},
        'properties': {'default': {}},
        'endpoint': {},
        'external': {},
        'level': {
            'type': 'integer',
            'choices': Level,
            'inherit': 'external.resource.level',
        },
        'access': {
            'type': 'string',
            'choices': Access,
            'inherit': 'external.resource.access',
            'default': 'protected',
        },
        'lang': {'type': 'object'},
        'params': {'type': 'object'},
        'comments': {},
    }

    def __init__(self):
        super().__init__()
        self.unique = []
        self.extends = None
        self.keymap = None
        self.backend = None
        self.version = None
        self.date = None
        self.link = None
        self.properties = {}
        self.flatprops = {}
        self.leafprops = {}
        self.given = ModelGiven()
        self.params = {}

    def model_type(self):
        return self.name

    def get_name_without_ns(self):
        if '/' in self.name:
            return self.name.split('/')[-1]
        else:
            return self.name


class PropertyGiven:
    access: str = None
    enum: str = None
    unit: str = None


class Property(Node):
    place: str = None  # Dotted property path
    title: str = None
    description: str = None
    link: str = None
    hidden: bool = False
    access: Access
    level: Level
    dtype: DataType = None
    external: Attribute
    list: Property = None
    model: Model = None
    uri: str = None
    given: PropertyGiven
    enum: EnumValue = None        # Enum name from Enums dict.
    enums: Enums
    lang: LangData = None
    unit: Unit = None       # Given in ref column.
    comments: List[Comment] = None

    schema = {
        'title': {},
        'description': {},
        'link': {},
        'hidden': {'type': 'boolean', 'default': False},
        'level': {
            'type': 'integer',
            'choices': Level,
            'inherit': 'model.level',
        },
        'access': {
            'type': 'string',
            'choices': Access,
            'inherit': 'model.access',
            'default': 'protected',
        },
        'external': {},
        'uri': {'type': 'string'},
        'enum': {'type': 'string'},
        'enums': {},
        'units': {'type': 'string'},
        'lang': {'type': 'object'},
        'comments': {},
    }

    def __init__(self):
        self.given = PropertyGiven()

    def __repr__(self):
        pypath = [type(self).__module__, type(self).__name__]
        pypath = '.'.join(pypath)
        dtype = self.dtype.name if self.dtype else 'none'
        kwargs = [
            f'name={self.place!r}',
            f'type={dtype!r}',
            f'model={self.model.name!r}',
        ]
        kwargs = ', '.join(kwargs)
        return f'<{pypath}({kwargs})>'

    def model_type(self):
        return f'{self.model.name}.{self.place}'

    def is_reserved(self):
        return self.name.startswith('_')


class Command:

    def __init__(self):
        self.name = None
        self.command = None
        self.args = None

    def __call__(self, *args, **kwargs):
        return self.command(*args, **self.args, **kwargs)


class CommandList:

    def __init__(self):
        self.commands = None

    def __call__(self, *args, **kwargs):
        return [command(*args, **kwargs) for command in self.commands]


@dataclasses.dataclass
class Attachment:
    content_type: str
    filename: str
    data: bytes


class Action(enum.Enum):
    INSERT = 'insert'
    UPSERT = 'upsert'
    UPDATE = 'update'
    PATCH = 'patch'
    DELETE = 'delete'

    WIPE = 'wipe'

    GETONE = 'getone'
    GETALL = 'getall'
    SEARCH = 'search'

    CHANGES = 'changes'

    CHECK = 'check'

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_

    @classmethod
    def by_value(cls, value):
        return cls._value2member_map_[value]

    @classmethod
    def values(cls):
        return list(cls._value2member_map_.keys())


class UrlParseNode(TypedDict):
    name: str
    args: List[Any]


class UrlParams:
    parsetree: List[UrlParseNode]

    path_parts: List[str] = None
    path: Optional[str] = None
    model: Optional[Model] = None
    pk: Optional[str] = None
    prop: Optional[Property] = None
    # Tells if we accessing property content or reference. Applies only for some
    # property types, like references.
    propref: bool = False

    # List only models names
    ns: bool = False
    # Recursively select all models in a given namespace
    all: bool = False
    external: bool = False

    changes: bool = False
    changes_offset: Optional[str] = None

    fmt: Format
    format: Optional[str] = None
    formatparams: dict

    select: Optional[List[str]] = None
    sort: List[dict] = None
    limit: Optional[int] = None
    offset: Optional[int] = None
    # Limit can be enforced even if it is not explicitly given in URL.
    limit_enforced: bool = False
    limit_enforced_to: int = 100
    # XXX: Deprecated, count should be part of `select`.
    count: bool = False
    # In batch requests, return summary of what was done.
    summary: bool = False
    # In batch requests, continue execution even if some actions fail.
    fault_tolerant: bool = False

    action: Action = None

    # If True, then HTTP HEAD request was made, this means no data must be
    # returned only headers.
    head: bool = False

    query: List[Dict[str, Any]] = None

    def changed_parsetree(self, change):
        ptree = {x['name']: x['args'] for x in (self.parsetree or [])}
        ptree.update(change)
        return [
            {'name': k, 'args': v} for k, v in ptree.items()
        ]


class Version:
    version: str


class DataItem:
    model: Optional[Model] = None       # Data model.
    prop: Optional[Property] = None     # Action on a property, not a whole model.
    propref: bool = False               # Action on property reference or instance.
    backend: Optional[Backend] = None   # Model or property backend depending on prop and propref.
    action: Optional[Action] = None     # Action.
    payload: Optional[dict] = None      # Original data from request.
    given: Optional[dict] = None        # Request data converted to Python-native data types.
    saved: Optional[dict] = None        # Current data stored in database.
    patch: Optional[dict] = None        # Patch that is going to be stored to database.
    error: Optional[exceptions.UserError] = None  # Error while processing data.

    def __init__(
        self,
        model: Optional[Model] = None,
        prop: Optional[Property] = None,
        propref: bool = False,
        backend: Optional[Backend] = None,
        action: Optional[Action] = None,
        payload: Optional[dict] = None,
        error: Optional[exceptions.UserError] = None,
    ):
        self.model = model
        self.prop = prop
        self.propref = propref
        self.backend = model.backend if backend is None and model else backend
        self.action = action
        self.payload = payload
        self.error = error
        self.given = NA
        self.saved = NA
        self.patch = NA

    def __getitem__(self, key):
        return DataSubItem(self, *(
            d.get(key, NA) if d else NA
            for d in (self.given, self.saved, self.patch)
        ))

    def copy(self, **kwargs) -> 'DataItem':
        data = DataItem()
        attrs = [
            'model',
            'prop',
            'propref',
            'backend',
            'action',
            'payload',
            'given',
            'saved',
            'patch',
            'error',
        ]
        assert len(set(kwargs) - set(attrs)) == 0
        for name in attrs:
            if name in kwargs:
                setattr(data, name, kwargs[name])
            elif hasattr(self, name):
                value = getattr(self, name)
                if isinstance(value, dict):
                    setattr(data, name, value.copy())
                else:
                    setattr(data, name, value)
        return data


class DataSubItem:

    def __init__(self, parent, given, saved, patch):
        if isinstance(parent, DataSubItem):
            self.root = parent.root
        else:
            self.root = parent
        self.given = given
        self.saved = saved
        self.patch = patch

    def __getitem__(self, key):
        return DataSubItem(self, *(
            d.get(key, NA) if d else NA
            for d in (self.given, self.saved, self.patch)
        ))

    def __iter__(self):
        yield from self.iter(given=True, saved=True, patch=True)

    def iter(self, given=False, saved=False, patch=False):
        if saved and self.saved:
            given_ = NA
            patch_ = NA
            for saved_ in self.saved:
                yield DataSubItem(self, given_, saved_, patch_)

        if (patch or given) and self.given and self.patch:
            saved_ = NA
            for given_, patch_ in zip(self.given, self.patch):
                yield DataSubItem(self, given_, saved_, patch_)

        elif given and self.given:
            saved_ = NA
            patch_ = NA
            for given_ in self.given:
                yield DataSubItem(self, given_, saved_, patch_)


DataStream = AsyncIterator[DataItem]


class Mode(enum.Enum):
    # Internal mode always use internal backend set on manifest, namespace or
    # model.
    internal = 'internal'

    # External model always sue external backend set on dataset or model's
    # source entity.
    external = 'external'


ScopeFormatterFunc = Callable[[
    Context,
    Union[Namespace, Model, Property],
    Action,
], str]


class Config:
    """Spinta configuration

    This is the place, where all Spinta configuration options are stored and
    used at runtime.

    """
    # TODO: `rc` should not be here, because `Config` might be initialized from
    #       different configuration sources.
    rc: RawConfig
    debug: bool = False
    config_path: pathlib.Path
    server_url: str
    scope_prefix: str
    scope_formatter: ScopeFormatterFunc
    scope_max_length: int
    default_auth_client: str
    http_basic_auth: bool
    token_validation_key: dict = None
    datasets: dict
    env: str
    docs_path: pathlib.Path
    always_show_id: bool = False
    # Limit access to specified namespace root.
    root: str = None
    credentials_file: pathlib.Path
    data_path: pathlib.Path
    AccessLog: Type[AccessLog]
    exporters: Dict[str, Format]

    def __init__(self):
        self.commands = _CommandsConfig()
        self.components = {}
        self.exporters = {}
        self.backends = {}
        self.manifests = {}
        self.ignore = []
        self.debug = False



from __future__ import annotations

import functools
from typing import List
from typing import Optional
from typing import TypedDict

import lark

from spinta.utils.schema import NA

GRAMMAR = r'''
?start: testlist
?testlist: test ("," test)* [","]
?test: or
?or: and ("|" and)*
?and: not ("&" not)*
?not: "!" not | comp
?comp: expr (COMP expr)*
?expr: term (TERM term)*
?term: factor (FACTOR factor)*
?factor: SIGN factor | composition
?composition: atom trailer*
?atom: "(" group? ")" | "[" list? "]" | func | value | name
group: test ("," test)* [","]
list: test ("," test)* [","]
?trailer: "[" filter? "]" | method | attr
func: NAME call
method: "." NAME call
?call: "(" arglist? ")"
arglist: argument ("," argument)*  [","]
?argument: test | kwarg
kwarg: NAME ":" test
filter: test ("," test)* [","]
attr: "." NAME
name: NAME
value: NULL | BOOL | INT | FLOAT | STRING | ALL

COMP: ">=" | "<=" | "!=" | "=" | "<" | ">"
TERM: "+" | "-"
FACTOR: "*" | "/" | "%"
SIGN: "+" | "-"

NAME: /[a-z_][a-z0-9_]*/i

ALL: "*"
STRING: /"(?!"").*?(?<!\\)(\\\\)*?"|'(?!'').*?(?<!\\)(\\\\)*?'/i
FLOAT: /(\d+)?\.\d+/
INT: /0|[1-9]\d*/
BOOL: "false" | "true"
NULL: "null"

COMMENT: /#[^\n]*/
WS: /[ \t\f\r\n]+/

%ignore WS
%ignore COMMENT
'''

_parser = lark.Lark(GRAMMAR, parser='lalr')


class SpynaAST(TypedDict):
    name: str
    args: List[SpynaAST]


def parse(rql) -> Optional[SpynaAST]:
    if not isinstance(rql, str):
        return rql

    if not rql:
        return None

    if rql.startswith('_id=') and len(rql) == 42:
        # Performance optimization for
        # _id='d6ee6808-6fc8-43eb-b1a5-cfb21c4be906' case.
        _id = rql[5:41]

    elif rql.startswith('eq(_id,') and len(rql) == 47:
        # Performance optimization for
        # eq(_id, 'd6ee6808-6fc8-43eb-b1a5-cfb21c4be906') case.
        _id = rql[9:45]

    else:
        ast = _parser.parse(rql)
        visit = Visitor()
        return visit(ast)

    return {'name': 'eq', 'args': [{'name': 'bind', 'args': ['_id']}, _id]}


class Visitor:

    def __call__(self, node):
        if isinstance(node, lark.Tree):
            return getattr(self, node.data, self.__default__)(node, *node.children)
        elif isinstance(node, lark.Token):
            return getattr(self, node.type)(node)
        else:
            return node

    def __default__(self, node, *args):
        if node.data in ('and', 'or'):
            return {
                'type': 'expression',
                'name': node.data,
                'args': self._args(*args),
            }
        else:
            return {
                'name': node.data,
                'args': self._args(*args),
            }

    def _name(self, node):
        assert node.data == 'name', node
        return [c.value for c in node.children]

    def _args(self, *args):
        return [self(arg) for arg in args]

    def comp(self, node, left, op, right):
        ops = {
            '=': 'eq',
            '!=': 'ne',
            '<': 'lt',
            '<=': 'le',
            '>': 'gt',
            '>=': 'ge',
        }
        return {
            'type': 'expression',
            'name': ops[op.value],
            'args': self._args(left, right),
        }

    def name(self, node, token):
        if token.value in ('null', 'false', 'true'):
            return self._const(token.value)
        return {
            'name': 'bind',
            'args': [token.value],
        }

    def attr(self, node, name):
        return {
            'name': 'bind',
            'args': [name.value],
        }

    def NAME(self, token):
        return {
            'name': 'bind',
            'args': [token.value],
        }

    def kwarg(self, node, name, value):
        return {
            'name': 'bind',
            'args': [name.value, self(value)],
        }

    def value(self, node, token):
        if token.type == 'STRING':
            return token.value[1:-1]
        if token.type == 'INT':
            return int(token.value)
        if token.type == 'FLOAT':
            return float(token.value)
        if token.type == 'NULL':
            return None
        if token.type == 'BOOL':
            return self._const(token.value)
        if token.type == 'ALL':
            return {
                'name': 'op',
                'args': ['*'],
            }
        raise Exception(f"Unknown token type: {token.type}")

    def _const(self, name: str):
        return {
            'null': None,
            'false': False,
            'true': True,
        }[name]

    def func(self, node, name, args):
        return {
            'name': name.value,
            'args': self._args(*args.children),
        }

    def method_comp(self, node, arg, name, args):
        return {
            'type': 'method',
            'name': name.value,
            'args': self._args(arg, *args.children),
        }

    def attr_comp(self, node, arg, name):
        return {
            'name': 'getattr',
            'args': self._args(arg, name),
        }

    def filter_comp(self, node, arg, *args):
        return {
            'name': 'filter',
            'args': self._args(arg, self._args(*args)),
        }

    def composition(self, node, *args):
        res = None
        for arg in args:
            if res is None:
                res = arg
            else:
                handler = getattr(self, arg.data + '_comp')
                res = handler(arg, res, *arg.children)
        return res

    def factor(self, node, sign, expr):
        names = {
            '+': 'positive',
            '-': 'negative',
        }
        return {
            'name': names[sign.value],
            'args': [self(expr)],
        }

    def expr(self, node, *args):
        names = {
            '+': 'add',
            '-': 'sub',
        }
        args = iter(args)
        left = self(next(args))
        for term in args:
            right = self(next(args))
            left = {
                'name': names[term.value],
                'args': [left, right],
            }
        return left

    def term(self, node, *args):
        names = {
            '*': 'mul',
            '/': 'div',
            '%': 'mod',
        }
        args = iter(args)
        left = self(next(args))
        for term in args:
            right = self(next(args))
            left = {
                'name': names[term.value],
                'args': [left, right],
            }
        return left


def unparse(rql, *, pretty=False, raw=False):
    if rql is NA:
        return '<NA>'
    if rql is None:
        return 'null'
    if rql is True:
        return 'true'
    if rql is False:
        return 'false'
    if not isinstance(rql, dict) or not set(rql) >= {'name', 'args'}:
        return repr(rql)

    ops = {
        'eq': '=',
        'ne': '!=',
        'lt': '<',
        'le': '<=',
        'gt': '>',
        'ge': '>=',
        'and': '&',
        'or': '|',
    }

    typ = rql.get('type')
    name = rql['name']
    _unparse = functools.partial(unparse, raw=raw)

    if not raw:
        if typ == 'expression':
            op = ops[name]
            args = (_unparse(arg) for arg in rql['args'])
            return op.join(args)

        if name == 'bind':
            if len(rql['args']) == 2:
                bind, value = rql['args']
                return bind + ': ' + _unparse(value)
            else:
                bind, = rql['args']
                return bind

        if name == 'getattr':
            obj, key = rql['args']
            return _unparse(obj) + '.' + _unparse(key)

        if name == 'filter':
            obj, group = rql['args']
            return _unparse(obj) + '[' + ', '.join(_unparse(arg) for arg in group) + ']'

        if name == 'positive':
            arg, = rql['args']
            return '+' + _unparse(arg)

        if name == 'negative':
            arg, = rql['args']
            return '-' + _unparse(arg)

        if name == 'testlist':
            return ', '.join(_unparse(arg) for arg in rql['args'])

        if name == 'group':
            return '(' + ', '.join(_unparse(arg) for arg in rql['args']) + ')'

        if name == 'list':
            return '[' + ', '.join(_unparse(arg) for arg in rql['args']) + ']'

        if name == 'op':
            return rql['args'][0]

        if name in ('add', 'sub', 'mul', 'div', 'mod'):
            symbols = {
                'add': '+',
                'sub': '-',
                'mul': '*',
                'div': '/',
                'mod': '%',
            }
            symbol = symbols[name]
            left, right = rql['args']
            return _unparse(left) + f' {symbol} ' + _unparse(right)

    args = [_unparse(arg) for arg in rql['args']]
    if not raw and typ == 'method':
        attr, args = args[0], args[1:]
        name = f'{attr}.{name}'
    sig = ', '.join(args)
    if pretty and len(sig) > 42 and len(args) > 1:
        sig = '\n    ' + ',\n    '.join(args) + '\n'
    return f'{name}({sig})'



from typing import Iterator
from typing import List
from typing import Optional

from spinta.backends import SelectTree
from spinta.backends import get_model_reserved_props
from spinta.backends.helpers import get_ns_reserved_props
from spinta.backends.helpers import get_select_prop_names
from spinta.backends.helpers import get_select_tree
from spinta.backends.helpers import select_only_props
from spinta.components import Action
from spinta.components import Context
from spinta.components import Model
from spinta.components import UrlParams
from spinta.types.datatype import Array
from spinta.types.datatype import DataType
from spinta.types.datatype import Object
from spinta.types.datatype import File
from spinta.types.datatype import Ref
from spinta.utils.data import take


def _get_dtype_header(
    dtype: DataType,
    select: SelectTree,
    name: str,
) -> Iterator[str]:
    if isinstance(dtype, Object):
        for prop, sel in select_only_props(
            dtype.prop,
            take(dtype.properties).keys(),
            dtype.properties,
            select,
        ):
            name_ = name + '.' + prop.name
            yield from _get_dtype_header(prop.dtype, sel, name_)

    elif isinstance(dtype, Array):
        name_ = name + '[]'
        yield from _get_dtype_header(dtype.items.dtype, select, name_)

    elif isinstance(dtype, File):
        yield name + '._id'
        yield name + '._content_type'

    elif isinstance(dtype, Ref):
        if select is None or select == {'*': {}}:
            yield name + '._id'
        else:
            for prop, sel in select_only_props(
                dtype.prop,
                dtype.model.properties.keys(),
                dtype.model.properties,
                select,
            ):
                name_ = name + '.' + prop.name
                yield from _get_dtype_header(prop.dtype, sel, name_)

    else:
        yield name


def _get_model_header(
    model: Model,
    names: List[str],
    select: SelectTree,
    reserved: List[str],
) -> List[str]:
    if select is None or select == {'*': {}}:
        keys = reserved + names
    else:
        keys = names
    props = model.properties
    props_ = select_only_props(
        model,
        keys,
        props,
        select,
        reserved=True,
    )
    for prop, sel in props_:
        yield from _get_dtype_header(prop.dtype, sel, prop.name)


def get_model_tabular_header(
    context: Context,
    model: Model,
    action: Action,
    params: UrlParams,
    *,
    reserved: Optional[List[str]] = None,
) -> List[str]:
    if params.count:
        header = ['count()']
    else:
        if reserved is None:
            if model.name == '_ns':
                reserved = get_ns_reserved_props(action)
            else:
                reserved = get_model_reserved_props(action)
        select = get_select_tree(context, action, params.select)
        if model.name == '_ns':
            names = get_select_prop_names(
                context,
                model,
                model.properties,
                action,
                select,
                auth=False,
            )
        else:
            names = get_select_prop_names(
                context,
                model,
                model.properties,
                action,
                select,
                reserved=reserved,
            )
        header = list(_get_model_header(model, names, select, reserved))
    return header



BYTE_MULTIPLES = {
    'b': 1,
    'k': 1000,
    'm': 1000**2,
    'g': 1000**3,
    't': 1000**4,
    'p': 1000**5,
    'e': 1000**6,
    'z': 1000**7,
    'y': 1000**8,
}


def tobytes(s, default: str = 'b'):
    unit = s[-1].lower()
    if unit in BYTE_MULTIPLES:
        s = s[:-1]
    else:
        unit = default
    value = int(s)
    return value * BYTE_MULTIPLES[unit]


TIME_UNITS = {
    's': 1,
    'm': 60,
    'h': 60 * 60,
    'd': 60 * 60 * 24,
    'w': 60 * 60 * 24 * 7,
    'n': 60 * 60 * 24 * 30,
    'y': 60 * 60 * 24 * 365,
}


def toseconds(s, default: str = 's'):
    unit = s[-1].lower()
    if unit in TIME_UNITS:
        s = s[:-1]
    else:
        unit = default
    value = int(s)
    return value * TIME_UNITS[unit]



import contextlib

import pymongo

from spinta.backends.components import BackendFeatures
from spinta.components import Model
from spinta.components import Property
from spinta.backends.components import Backend


class Mongo(Backend):
    # Instance of this class will be created when application starts. First if
    # will be loaded from configuration using `load` command, then it will be
    # prepared from manifest declarations using `prepare` command.
    #
    # Backend also must have a `transaction` method which must return read or
    # write transaction object containing an active `connection` to database.
    metadata = {
        'name': 'mongo',
        'properties': {
            'dsn': {'type': 'string', 'required': True},
            'db': {'type': 'string', 'required': True},
        },
    }

    features = {
        BackendFeatures.WRITE,
    }

    @contextlib.contextmanager
    def transaction(self, write=False):
        with self.begin() as connection:
            if write:
                # TODO: get a real transaction id
                transaction_id = 1
                yield WriteTransaction(connection, transaction_id)
            else:
                yield ReadTransaction(connection)

    @contextlib.contextmanager
    def begin(self):
        yield self

    def get_table(
        self,
        model: Model,
        name: str = None,
    ) -> pymongo.collection.Collection:
        return self.db[model.model_type()]

    def get_column(
        self,
        table: pymongo.collection.Collection,
        prop: Property,
        *,
        select=False,
    ) -> str:
        return prop.place


class ReadTransaction:

    def __init__(self, connection):
        self.connection = connection


class WriteTransaction(ReadTransaction):

    def __init__(self, connection, id):
        super().__init__(connection)
        self.id = id
        self.errors = 0



import datetime
import uuid
from typing import Dict

import contextlib

from spinta.typing import ObjectData
from spinta.backends.components import Backend
from spinta.backends.components import BackendFeatures


class Memory(Backend):
    metadata = {
        'name': 'memory',
    }

    features = {
        BackendFeatures.WRITE,
    }

    data: Dict[
        str,                # table
        Dict[
            str,            # id
            ObjectData,     # data
        ]
    ]

    def __init__(self):
        self.data = {}

    @contextlib.contextmanager
    def transaction(self, write=False):
        if write:
            obj = self.insert({
                '_type': '_txn',
                'datetime': datetime.datetime.utcnow(),
                'client_type': '',
                'client_id': '',
                'errors': 0,
            })
            yield WriteTransaction(self, obj['_id'])
        else:
            yield ReadTransaction(self)

    @contextlib.contextmanager
    def begin(self):
        yield self

    def bootstrapped(self):
        return True

    def create(self, table: str):
        if table not in self.data:
            self.data[table] = {}

    def insert(self, obj: ObjectData):
        obj = obj.copy()

        table = obj['_type']
        if table not in self.data:
            raise RuntimeError(f"Table {table!r} does not exist.")

        if '_id' not in obj:
            obj['_id'] = str(uuid.uuid4())

        pk = obj['_id']
        self.data[table][pk] = obj

        return obj


class ReadTransaction:
    id: str
    errors: int
    connection: Memory

    def __init__(self, connection: Memory):
        self.connection = connection


class WriteTransaction(ReadTransaction):

    def __init__(self, connection: Memory, id_: str):
        super().__init__(connection)
        self.id = id_
        self.errors = 0



from typing import Optional

from spinta.components import Component


class KeyMap(Component):
    name: str = None

    def encode(self, name: str, value: object) -> Optional[str]:
        raise NotImplementedError

    def decode(self, name: str, key: str) -> object:
        raise NotImplementedError



from typing import Optional

from typer import Argument

from spinta.manifests.tabular.helpers import write_tabular_manifest


def init(
    manifest: Optional[str] = Argument(None, help="path to a manifest"),
):
    """Initialize a new manifest table

    `MANIFEST` is a file path, for example `manifest.csv` or `manifest.xlsx`.
    Depending of file extensions, a CSV of a XLSX format manifest will be
    created.
    """
    write_tabular_manifest(manifest)




from typing import Dict, Any

import json

from spinta import commands
from spinta.accesslog import AccessLog
from spinta.components import Context, Store


class PythonAccessLog(AccessLog):
    stream: list

    def log(self, message: Dict[str, Any]) -> None:
        # `message` must be JSON serializable.
        json.dumps(message)
        self.stream.append(message)


@commands.load.register(Context, PythonAccessLog, Store)
def load(context: Context, accesslog: PythonAccessLog, store: Store):
    if context.has('accesslog.stream'):
        accesslog.stream = context.get('accesslog.stream')
    else:
        accesslog.stream = []



from typing import Optional

from starlette.requests import Request
from starlette.responses import JSONResponse, Response, StreamingResponse

from spinta.components import Context, Action, UrlParams, Node
from spinta import commands
from spinta.formats.json.components import Json
from spinta.utils.response import aiter, peek_and_stream


@commands.render.register(Context, Request, Node, Json)
def render(
    context: Context,
    request: Request,
    node: Node,
    fmt: Json,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    return _render(
        context,
        request,
        fmt,
        action,
        params,
        data,
        status_code,
        headers,
    )


def _render(
    context: Context,
    request: Request,
    fmt: Json,
    action: Action,
    params: UrlParams,
    data,
    status_code: int,
    headers: dict,
):
    if action in (Action.GETALL, Action.SEARCH, Action.CHANGES):
        # In python dict is also an iterable, but here we want a true iterable,
        # a list or a generator of dicts.
        assert not isinstance(data, dict), data
        return StreamingResponse(
            aiter(peek_and_stream(fmt(data))),
            status_code=status_code,
            media_type=fmt.content_type,
            headers=headers,
        )
    elif action == Action.DELETE:
        return Response(None, status_code=status_code, headers=headers)
    else:
        return JSONResponse(
            fmt.data(data),
            status_code=status_code,
            headers=headers
        )



from typing import Union, List

from spinta import exceptions


class NotAvailable:

    def __repr__(self):
        return "<NA>"

    def __bool__(self):
        return False


NA = NotAvailable()


def resolve_schema(obj, Base):
    # Emulates inheritance for `schema` class attribute for given `obj` and
    # one of the `obj's` base classes `Base`.
    #
    # e.g.:
    #
    # class Base:
    #     schema = { 'required': {'type': 'bool', 'default': False} }
    #
    # class Obj(Base):
    #     schema = { 'items': {} }
    #
    # obj = Obj()
    # full_schema = resolve_schema(obj, Base)
    # print(full_schema)
    # { 'required': {'type': 'bool', 'default': False}, 'items': {} }

    bases = []
    for cls in obj.__class__.mro():
        bases.append(cls)
        if cls is Base:
            break
    else:
        raise Exception(f"Could not find specified base {Base!r} on {obj.__class__}.")
    bases = reversed(bases)
    schema = {}
    for cls in bases:
        if hasattr(cls, 'schema'):
            schema.update(cls.schema)
    return schema


def load_from_schema(Base: type, obj: object, params: dict, check_unknowns=True):
    schema = resolve_schema(obj, Base)
    for name in set(schema) | set(params):
        if name not in schema:
            if check_unknowns:
                raise exceptions.UnknownParameter(obj, param=name)
            else:
                continue
        value = params.get(name, NA)
        value = _get_value(obj, schema[name], name, value)
        setattr(obj, name, value)
    return obj


def _get_value(obj: object, schema: dict, name: str, value: object):
    if schema.get('required', False) and value is NA:
        raise Exception(f"Missing required param {name!r}.")
    if value is NA:
        value = schema.get('default')
    return value


def check_unkown_params(
    schema: Union[List[dict], dict],
    data: dict, node,
):
    schemas = schema if isinstance(schema, list) else [schema]
    known_params = set.union(*(set(s.keys()) for s in schemas))
    given_params = set(data.keys())
    unknown_params = given_params - known_params
    if unknown_params:
        raise exceptions.MultipleErrors(
            exceptions.UnknownParameter(node, param=param)
            for param in sorted(unknown_params)
        )


def is_valid_sort_key(key, model):
    # sort key must be a leaf node in the model schema.
    # we cannot sort using intermediate node, because it's type would
    # be `array` or `object`.
    #
    # is_valid_sort_key('certificates', report_model) == False
    # is_valid_sort_key('certificates.notes.note_type', report_model) == True
    leaf_key = key.split('.')[-1]
    if leaf_key not in model.leafprops:
        return False
    return True



from typing import Iterator
from typing import overload

from multipledispatch import dispatch

from spinta.typing import ObjectData
from spinta import commands
from spinta.components import Context
from spinta.components import Model
from spinta.components import Property
from spinta.types.datatype import DataType
from spinta.types.datatype import File
from spinta.types.datatype import Object
from spinta.exceptions import NotFoundError, ItemDoesNotExist
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name
from spinta.utils.nestedstruct import flat_dicts_to_nested


@overload
@commands.getone.register(Context, Property, Object, PostgreSQL)
def getone(
    context: Context,
    prop: Property,
    dtype: Object,
    backend: PostgreSQL,
    *,
    id_: str,
) -> ObjectData:
    table = backend.get_table(prop.model)
    connection = context.get('transaction').connection
    selectlist = [
        table.c._id,
        table.c._revision,
    ] + [
        table.c[name]
        for name in _iter_prop_names(prop.dtype)
    ]
    try:
        data = backend.get(connection, selectlist, table.c._id == id_)
    except NotFoundError:
        raise ItemDoesNotExist(prop.model, id=id_)

    result = {
        '_type': prop.model_type(),
        '_id': data[table.c._id],
        '_revision': data[table.c._revision],
    }

    data = flat_dicts_to_nested(data)
    result[prop.name] = data[prop.name]
    return commands.cast_backend_to_python(context, prop, backend, result)


@dispatch((Model, Object))
def _iter_prop_names(dtype) -> Iterator[Property]:
    for prop in dtype.properties.values():
        yield from _iter_prop_names(prop.dtype)


@dispatch(DataType)
def _iter_prop_names(dtype) -> Iterator[Property]:  # noqa
    if not dtype.prop.name.startswith('_'):
        yield get_column_name(dtype.prop)


@dispatch(File)
def _iter_prop_names(dtype) -> Iterator[Property]:  # noqa
    yield dtype.prop.place + '._id'
    yield dtype.prop.place + '._content_type'



from spinta import commands
from spinta.components import Context
from spinta.manifests.backend.components import BackendManifest


@commands.configure.register(Context, BackendManifest)
def configure(context: Context, manifest: BackendManifest):
    pass



from spinta.utils.schema import NA
from spinta.components import DataSubItem, Action


def inserting(data: DataSubItem):
    return (
        data.root.action == Action.INSERT or
        (data.root.action == Action.UPSERT and data.saved is NA)
    )



from typing import Optional
from typing import Tuple

from typer import Argument
from typer import Context as TyperContext
from typer import Option
from typer import echo

from spinta import commands
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.store import prepare_manifest
from spinta.components import Mode
from spinta.core.context import configure_context
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import init_manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.tabular.helpers import render_tabular_manifest
from spinta.manifests.tabular.helpers import write_tabular_manifest
from spinta.core.config import parse_resource_args


def inspect(
    ctx: TyperContext,
    manifest: Optional[str] = Argument(None, help="Path to manifest."),
    resource: Optional[Tuple[str, str]] = Option((None, None), '-r', '--resource', help=(
        "Resource type and source URI (-r sql sqlite:////tmp/db.sqlite)"
    )),
    formula: str = Option('', '-f', '--formula', help=(
        "Formula if needed, to prepare resource for reading"
    )),
    backend: Optional[str] = Option(None, '-b', '--backend', help=(
        "Backend connection string"
    )),
    output: Optional[str] = Option(None, '-o', '--output', help=(
        "Output tabular manifest in a specified file"
    )),
    auth: Optional[str] = Option(None, '-a', '--auth', help=(
        "Authorize as a client"
    )),
):
    """Update manifest schema from an external data source"""
    resources = parse_resource_args(*resource, formula)
    context = configure_context(
        ctx.obj,
        [manifest] if manifest else None,
        mode=Mode.external,
        backend=backend,
        resources=resources,
    )
    store = prepare_manifest(context, ensure_config_dir=True)

    manifest = Manifest()
    with context:
        require_auth(context, auth)
        init_manifest(context, manifest, 'inspect')
        manifest.keymap = store.manifest.keymap
        schemas = commands.inspect(
            context,
            store.manifest.backend,
            store.manifest,
            None,
        )
        load_manifest_nodes(context, manifest, schemas)
        commands.link(context, manifest)

    if output:
        write_tabular_manifest(output, manifest)
    else:
        echo(render_tabular_manifest(manifest))



from typing import Any
from typing import Dict
from typing import overload

from geoalchemy2 import WKBElement
from geoalchemy2.shape import to_shape


from spinta import commands
from spinta.backends.postgresql.types.geometry.helpers import get_display_value
from spinta.backends.postgresql.types.geometry.helpers import get_osm_link
from spinta.formats.components import Format
from spinta.components import Action
from spinta.components import Context
from spinta.formats.html.components import Cell
from spinta.formats.html.components import Html
from spinta.types.geometry.components import Geometry


@overload
@commands.prepare_dtype_for_response.register(Context, Format, Geometry, WKBElement)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Geometry,
    value: WKBElement,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    shape = to_shape(value)
    return shape.wkt


@overload
@commands.prepare_dtype_for_response.register(Context, Html, Geometry, WKBElement)
def prepare_dtype_for_response(
    context: Context,
    fmt: Html,
    dtype: Geometry,
    value: WKBElement,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    shape = to_shape(value)
    display_value = get_display_value(shape)
    osm_link = get_osm_link(value, dtype.srid)
    return Cell(display_value, link=osm_link)



from __future__ import annotations

import base64
import dataclasses
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import TypedDict
from typing import Union

import sqlalchemy as sa
from sqlalchemy.sql.functions import Function

from spinta.auth import authorized
from spinta.components import Action
from spinta.components import Model
from spinta.components import Property
from spinta.core.ufuncs import Bind
from spinta.core.ufuncs import Env
from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import Negative
from spinta.core.ufuncs import ufunc
from spinta.datasets.backends.sql.components import Sql
from spinta.datasets.backends.sql.ufuncs.components import SqlResultBuilder
from spinta.dimensions.enum.helpers import prepare_enum_value
from spinta.dimensions.param.components import ResolvedParams
from spinta.exceptions import PropertyNotFound
from spinta.exceptions import UnknownMethod
from spinta.types.datatype import DataType
from spinta.types.datatype import PrimaryKey
from spinta.types.datatype import Ref
from spinta.types.datatype import String
from spinta.types.file.components import FileData
from spinta.ufuncs.components import ForeignProperty
from spinta.core.ufuncs import Unresolved
from spinta.utils.data import take
from spinta.utils.itertools import flatten
from spinta.utils.schema import NA


def ensure_list(value: Any) -> List[Any]:
    if isinstance(value, tuple):
        return list(value)
    if isinstance(value, list):
        return value
    else:
        return [value]


class SqlFrom:
    backend: Sql
    joins: Dict[str, sa.Table]
    from_: sa.Table

    def __init__(self, backend: Sql, table: sa.Table):
        self.backend = backend
        self.joins = {}
        self.from_ = table

    def get_table(
        self,
        env: SqlQueryBuilder,
        prop: ForeignProperty,
    ) -> sa.Table:
        fpr: Optional[ForeignProperty] = None
        for fpr in prop.chain:
            if fpr.name in self.joins:
                continue

            # Left table foreign keys
            lmodel = fpr.left.prop.model
            if len(fpr.chain) > 1:
                # Use table alias of previous join.
                ltable = self.joins[fpr.chain[-2].name]
            else:
                # Use the main table, without alias.
                ltable = self.backend.get_table(lmodel)
            lenv = env(model=lmodel, table=ltable)
            lfkeys = lenv.call('join_table_on', fpr.left.prop)
            lfkeys = ensure_list(lfkeys)

            # Right table primary keys
            rpkeys = []
            rmodel = fpr.right.prop.model
            rtable = self.backend.get_table(rmodel).alias()
            renv = env(model=rmodel, table=rtable)
            for rpk in fpr.left.refprops:
                rpkeys += ensure_list(
                    renv.call('join_table_on', rpk)
                )

            # Number of keys on both left and right must be equal.
            assert len(lfkeys) == len(rpkeys), (lfkeys, rpkeys)
            condition = []
            for lfk, rpk in zip(lfkeys, rpkeys):
                condition += [lfk == rpk]

            # Build `JOIN rtable ON (condition)`.
            assert len(condition) > 0
            if len(condition) == 1:
                condition = condition[0]
            else:
                condition = sa.and_(*condition)

            self.joins[fpr.name] = rtable
            self.from_ = self.from_.outerjoin(rtable, condition)

        return self.joins[fpr.name]


class SqlQueryBuilder(Env):
    backend: Sql
    model: Model
    table: sa.Table
    joins: SqlFrom
    columns: List[sa.Column]
    # `resolved` is used to map which prop.place properties are already
    # resolved, usually it maps to Selected, but different DataType's can return
    # different results.
    resolved: Dict[str, Selected]
    selected: Dict[str, Selected] = None
    params: ResolvedParams

    def init(self, backend: Sql, table: sa.Table):
        return self(
            backend=backend,
            table=table,
            columns=[],
            resolved={},
            selected=None,
            joins=SqlFrom(backend, table),
            sort=[],
            limit=None,
            offset=None,
        )

    def build(self, where):
        if self.selected is None:
            # If select list was not explicitly given by client, then select all
            # properties.
            self.call('select', Expr('select'))

        qry = sa.select(self.columns)
        qry = qry.select_from(self.joins.from_)

        if where is not None:
            qry = qry.where(where)

        if self.sort:
            qry = qry.order_by(*self.sort)

        if self.limit is not None:
            qry = qry.limit(self.limit)

        if self.offset is not None:
            qry = qry.offset(self.offset)

        return qry

    def execute(self, expr: Any):
        expr = self.call('_resolve_unresolved', expr)
        return super().execute(expr)

    def default_resolver(self, expr, *args, **kwargs):
        raise UnknownMethod(name=expr.name, expr=str(expr(*args, **kwargs)))

    def add_column(self, column: Union[sa.Column, Function]) -> int:
        """Returns position in select column list, which is stored in
        Selected.item.
        """
        assert isinstance(column, (sa.Column, Function)), column
        if column not in self.columns:
            self.columns.append(column)
        return self.columns.index(column)


class Selected:
    # Item index in select list.
    item: int = None
    # Model property if a property is selected.
    prop: Property = None
    # A value or an Expr for further processing on selected value.
    prep: Any = NA

    def __init__(
        self,
        item: int = None,
        prop: Property = None,
        # `prop` can be Expr or any other value.
        prep: Any = NA,
    ):
        self.item = item
        self.prop = prop
        self.prep = prep

    def __repr__(self):
        return self.debug()

    def debug(self, indent: str = ''):
        prop = self.prop.place if self.prop else 'None'
        if isinstance(self.prep, Selected):
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep=...)\n'
                   ) + self.prep.debug(indent + '  ')
        elif isinstance(self.prep, (tuple, list)):
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep={type(self.prep).__name__}...)\n'
            ) + ''.join([
                p.debug(indent + '- ')
                if isinstance(p, Selected)
                else str(p)
                for p in self.prep
            ])
        else:
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep={self.prep})\n'
            )


@dataclasses.dataclass
class GetAttr(Unresolved):
    obj: str
    name: Union[GetAttr, Bind]


@ufunc.resolver(SqlQueryBuilder, Bind, Bind, name='getattr')
def getattr_(env: SqlQueryBuilder, obj: Bind, attr: Bind):
    return GetAttr(obj.name, attr)


@ufunc.resolver(SqlQueryBuilder, Bind, GetAttr, name='getattr')
def getattr_(env: SqlQueryBuilder, obj: Bind, attr: GetAttr):
    return GetAttr(obj.name, attr)


@ufunc.resolver(SqlQueryBuilder, GetAttr)
def _resolve_getattr(
    env: SqlQueryBuilder,
    attr: GetAttr,
) -> ForeignProperty:
    prop = env.model.properties[attr.obj]
    return env.call('_resolve_getattr', prop.dtype, attr.name)


@ufunc.resolver(SqlQueryBuilder, Ref, GetAttr)
def _resolve_getattr(
    env: SqlQueryBuilder,
    dtype: Ref,
    attr: GetAttr,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.obj]
    fpr = ForeignProperty(None, dtype, prop.dtype)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(SqlQueryBuilder, Ref, Bind)
def _resolve_getattr(
    env: SqlQueryBuilder,
    dtype: Ref,
    attr: Bind,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype, prop.dtype)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, Ref, GetAttr)
def _resolve_getattr(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: Ref,
    attr: GetAttr,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.obj]
    fpr = fpr.push(prop)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, Ref, Bind)
def _resolve_getattr(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: Ref,
    attr: Bind,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.name]
    return fpr.push(prop)


@ufunc.resolver(SqlQueryBuilder, str, object, names=['eq', 'ne'])
def eq(env: SqlQueryBuilder, op: str, field: str, value: Any):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    return env.call(op, Bind(field), value)


def _sa_compare(op: str, column: sa.Column, value: Any):
    if op == 'eq':
        return column == value

    if op == 'ne':
        return column != value

    if op == 'lt':
        return column < value

    if op == 'le':
        return column <= value

    if op == 'gt':
        return column > value

    if op == 'ge':
        return column >= value

    if op == 'contains':
        return column.contains(value)

    if op == 'startswith':
        return column.startswith(value)

    raise NotImplementedError


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]


T = TypeVar('T')


def _prepare_value(prop: Property, value: T) -> Union[T, List[T]]:
    return prepare_enum_value(prop, value)


@ufunc.resolver(SqlQueryBuilder, Bind, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, field: Bind, value: Any):
    prop = env.model.properties[field.name]
    value = _prepare_value(prop, value)
    return env.call(op, prop.dtype, value)


@ufunc.resolver(SqlQueryBuilder, GetAttr, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, attr: GetAttr, value: Any):
    fpr: ForeignProperty = env.call('_resolve_getattr', attr)
    value = _prepare_value(fpr.right.prop, value)
    return env.call(op, fpr, fpr.right, value)


@ufunc.resolver(SqlQueryBuilder, Bind, list, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, field: Bind, value: List[Any]):
    prop = env.model.properties[field.name]
    value = list(flatten(_prepare_value(prop, v) for v in value))
    return env.call(op, prop.dtype, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, fpr: ForeignProperty, value: Any):
    value = _prepare_value(fpr.right.prop, value)
    return env.call(op, fpr, fpr.right, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, list, names=COMPARE)
def compare(
    env: SqlQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    value: List[Any],
):
    value = list(flatten(_prepare_value(fpr.right.prop, v) for v in value))
    return env.call(op, fpr, fpr.right, value)


@ufunc.resolver(SqlQueryBuilder, DataType, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, dtype: DataType, value: Any):
    column = env.backend.get_column(env.table, dtype.prop)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, object, names=COMPARE)
def compare(
    env: SqlQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: DataType,
    value: Any,
):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, Function, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, func: Function, value: Any):
    return _sa_compare(op, func, value)


@ufunc.resolver(SqlQueryBuilder, DataType, list)
def eq(env: SqlQueryBuilder, dtype: DataType, value: List[Any]):
    column = env.backend.get_column(env.table, dtype.prop)
    return column.in_(value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, list)
def eq(env: SqlQueryBuilder, fpr: ForeignProperty, dtype: DataType, value: list):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop)
    return column.in_(value)


@ufunc.resolver(SqlQueryBuilder, DataType, list)
def ne(env: SqlQueryBuilder, dtype: DataType, value: List[Any]):
    column = env.backend.get_column(env.table, dtype.prop)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, list)
def ne(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: List[Any],
):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, fpr.right.prop)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, list)
def ne(env: SqlQueryBuilder, fpr: ForeignProperty, value: List[Any]):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, fpr.right.prop)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, object)
def _resolve_unresolved(env: SqlQueryBuilder, value: Any) -> Any:
    if isinstance(value, Unresolved):
        raise ValueError(f"Unresolved value {value!r}.")
    else:
        return value


@ufunc.resolver(SqlQueryBuilder, Bind)
def _resolve_unresolved(env: SqlQueryBuilder, field: Bind) -> sa.Column:
    prop = env.model.flatprops.get(field.name)
    if prop:
        return env.backend.get_column(env.table, prop)
    else:
        raise PropertyNotFound(env.model, property=field.name)


@ufunc.resolver(SqlQueryBuilder, GetAttr)
def _resolve_unresolved(env: SqlQueryBuilder, attr: GetAttr) -> sa.Column:
    fpr = env.call('_resolve_getattr', attr)
    table = env.joins.get_table(env, fpr)
    dtype = fpr.right
    return env.backend.get_column(table, dtype.prop)


@ufunc.resolver(SqlQueryBuilder, Expr, name='and')
def and_(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    args = [
        env.call('_resolve_unresolved', arg)
        for arg in args
        if arg is not None
    ]
    if len(args) > 1:
        return sa.and_(*args)
    elif args:
        return args[0]


@ufunc.resolver(SqlQueryBuilder, Expr, name='or')
def or_(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    args = [
        env.call('_resolve_unresolved', arg)
        for arg in args
        if arg is not None
    ]
    if len(args) > 1:
        return sa.or_(*args)
    elif args:
        return args[0]


@ufunc.resolver(SqlQueryBuilder, Expr, name='list')
def list_(env: SqlQueryBuilder, expr: Expr) -> List[Any]:
    args, kwargs = expr.resolve(env)
    return list(args)


@ufunc.resolver(SqlQueryBuilder, Expr)
def testlist(env: SqlQueryBuilder, expr: Expr) -> Tuple[Any]:
    args, kwargs = expr.resolve(env)
    return tuple(args)


@ufunc.resolver(SqlQueryBuilder)
def count(env: SqlQueryBuilder):
    return sa.func.count()


@ufunc.resolver(SqlQueryBuilder, Expr)
def select(env: SqlQueryBuilder, expr: Expr):
    keys = [str(k) for k in expr.args]
    args, kwargs = expr.resolve(env)
    args = list(zip(keys, args)) + list(kwargs.items())

    if env.selected is not None:
        raise RuntimeError("`select` was already called.")

    env.selected = {}
    if args:
        for key, arg in args:
            env.selected[key] = env.call('select', arg)
    else:
        for prop in take(['_id', all], env.model.properties).values():
            if authorized(env.context, prop, Action.GETALL):
                env.selected[prop.place] = env.call('select', prop)

    if not env.columns:
        raise RuntimeError(
            f"{expr} didn't added anything to select list."
        )


@ufunc.resolver(SqlQueryBuilder, object)
def select(env: SqlQueryBuilder, value: Any) -> Selected:
    """For things like select(1, count())."""
    return Selected(item=env.add_column(value))


@ufunc.resolver(SqlQueryBuilder, Bind)
def select(env: SqlQueryBuilder, item: Bind, *, nested: bool = False):
    prop = _get_property_for_select(env, item.name, nested=nested)
    return env.call('select', prop)


@ufunc.resolver(SqlQueryBuilder, str)
def select(env: SqlQueryBuilder, item: str, *, nested: bool = False):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    prop = _get_property_for_select(env, item, nested=nested)
    return env.call('select', prop)


def _get_property_for_select(
    env: SqlQueryBuilder,
    name: str,
    *,
    nested: bool = False,
) -> Property:
    # TODO: `name` can refer to (in specified order):
    #       - var - a defined variable
    #       - param - a parameter if parametrization is used
    #       - item - an item of a dict or list
    #       - prop - a property
    #       Currently only `prop` is resolved.
    prop = env.model.flatprops.get(name)
    if prop and (
        # Check authorization only for top level properties in select list.
        # XXX: Not sure if nested is the right property to user, probably better
        #      option is to check if this call comes from a prepare context. But
        #      then how prepare context should be defined? Probably resolvers
        #      should be called with a different env class?
        #      tag:resolving_private_properties_in_prepare_context
        nested or
        authorized(env.context, prop, Action.SEARCH)
    ):
        return prop
    else:
        raise PropertyNotFound(env.model, property=name)


@ufunc.resolver(SqlQueryBuilder, Property)
def select(env: SqlQueryBuilder, prop: Property) -> Selected:
    if prop.place not in env.resolved:
        if isinstance(prop.external, list):
            # TODO: Should raise one of spinta.exceptions exception, with
            #       property as a context.
            raise ValueError("Source can't be a list, use prepare instead.")
        if prop.external.prepare is not NA:
            # If `prepare` formula is given, evaluate formula.
            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = prop.external.prepare
            # XXX: Maybe interpretation of prepare formula should be done under
            #      a different Env? This way, select resolvers would know when
            #      properties are resolved under a formula context and for
            #      example would be able to decide to not check permissions on
            #      properties.
            #      tag:resolving_private_properties_in_prepare_context
            result = env.call('select', prop.dtype, result)
        elif prop.external and prop.external.name:
            # If prepare is not given, then take value from `source`.
            result = env.call('select', prop.dtype)
        elif prop.is_reserved():
            # Reserved properties never have external source.
            result = env.call('select', prop.dtype)
        else:
            # If `source` is not given, return None.
            result = Selected(prop=prop, prep=None)
        assert isinstance(result, Selected), prop
        env.resolved[prop.place] = result
    return env.resolved[prop.place]


@ufunc.resolver(SqlQueryBuilder, DataType)
def select(env: SqlQueryBuilder, dtype: DataType) -> Selected:
    table = env.backend.get_table(env.model)
    column = env.backend.get_column(table, dtype.prop, select=True)
    return Selected(
        item=env.add_column(column),
        prop=dtype.prop,
    )


@ufunc.resolver(SqlQueryBuilder, DataType, object)
def select(env: SqlQueryBuilder, dtype: DataType, prep: Any) -> Selected:
    if isinstance(prep, str):
        # XXX: Backwards compatibility thing.
        #      str values are interpreted as Bind values and Bind values are
        #      assumed to be properties. So here we skip
        #      `env.call('select', prep)` and return `prep` as is.
        #      This should be eventually removed, once backwards compatibility
        #      for resolving strings as properties is removed.
        return Selected(prop=dtype.prop, prep=prep)
    elif isinstance(prep, Expr):
        # If `prepare` expression returns another expression, then this means,
        # it must be processed on values returned by query.
        prop = dtype.prop
        if prop.external and prop.external.name:
            sel = env.call('select', dtype)
            return Selected(item=sel.item, prop=sel.prop, prep=prep)
        else:
            return Selected(item=None, prop=prop, prep=prep)
    else:
        result = env.call('select', prep)
        return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(SqlQueryBuilder, PrimaryKey)
def select(
    env: SqlQueryBuilder,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        # If primary key is not specified use all properties to uniquely
        # identify row.
        pkeys = take(model.properties).values()

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', prop)
    else:
        result = [
            env.call('select', prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(SqlQueryBuilder, Ref, object)
def select(env: SqlQueryBuilder, dtype: Ref, prep: Any) -> Selected:
    fpr = ForeignProperty(None, dtype, dtype.model.properties['_id'].dtype)
    return Selected(
        prop=dtype.prop,
        prep=env.call('select', fpr, fpr.right.prop),
    )


@ufunc.resolver(SqlQueryBuilder, GetAttr)
def select(env: SqlQueryBuilder, attr: GetAttr) -> Selected:
    """For things like select(foo.bar.baz)."""
    fpr: ForeignProperty = env.call('_resolve_getattr', attr)
    return Selected(
        prop=fpr.right.prop,
        prep=env.call('select', fpr, fpr.right.prop),
    )


@ufunc.resolver(SqlQueryBuilder, ForeignProperty)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
) -> Selected:
    return env.call('select', fpr, fpr.right.prop)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, Property)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    prop: Property,
) -> Selected:
    resolved_key = fpr / prop
    if resolved_key not in env.resolved:
        if isinstance(prop.external, list):
            raise ValueError("Source can't be a list, use prepare instead.")
        if prop.external.prepare is not NA:
            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = prop.external.prepare
            result = env.call('select', fpr, prop.dtype, result)
        else:
            result = env.call('select', fpr, prop.dtype)
        assert isinstance(result, Selected), prop
        env.resolved[resolved_key] = result
    return env.resolved[resolved_key]


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
) -> Selected:
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop, select=True)
    return Selected(
        item=env.add_column(column),
        prop=dtype.prop,
    )


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, object)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    prep: Any,
) -> Selected:
    result = env.call('select', fpr, prep)
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, tuple)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', fpr, v) for v in prep)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, Bind)
def select(env: SqlQueryBuilder, fpr: ForeignProperty, item: Bind):
    model = fpr.right.prop.model
    prop = model.flatprops.get(item.name)
    if prop and authorized(env.context, prop, Action.SEARCH):
        return env.call('select', fpr, prop)
    else:
        raise PropertyNotFound(model, property=item.name)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, PrimaryKey)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        raise RuntimeError(
            f"Can't join {dtype.prop} on right table without primary key."
        )

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', fpr, prop)
    else:
        result = [
            env.call('select', fpr, prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(SqlQueryBuilder, list)
def select(
    env: SqlQueryBuilder,
    prep: List[Any],
) -> List[Any]:
    return [env.call('select', v) for v in prep]


@ufunc.resolver(SqlQueryBuilder, tuple)
def select(
    env: SqlQueryBuilder,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', v) for v in prep)


@ufunc.resolver(SqlQueryBuilder, dict)
def select(
    env: SqlQueryBuilder,
    prep: Dict[str, Any],
) -> Dict[str, Any]:
    # TODO: Add tests.
    return {k: env.call('select', v) for k, v in prep.items()}


@ufunc.resolver(SqlQueryBuilder, Property)
def join_table_on(env: SqlQueryBuilder, prop: Property) -> Any:
    if prop.external.prepare is not NA:
        if isinstance(prop.external.prepare, Expr):
            result = env.resolve(prop.external.prepare)
        else:
            result = prop.external.prepare
        return env.call('join_table_on', prop.dtype, result)
    else:
        return env.call('join_table_on', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, DataType)
def join_table_on(env: SqlQueryBuilder, dtype: DataType) -> Tuple[Any]:
    column = env.backend.get_column(env.table, dtype.prop)
    return column


@ufunc.resolver(SqlQueryBuilder, PrimaryKey)
def join_table_on(env: SqlQueryBuilder, dtype: DataType) -> Any:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        raise RuntimeError(
            f"Can't join {dtype.prop} on right table without primary key."
        )

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('join_table_on', prop)
    else:
        result = [
            env.call('join_table_on', prop)
            for prop in pkeys
        ]

    return result


@ufunc.resolver(SqlQueryBuilder, DataType, tuple)
def join_table_on(
    env: SqlQueryBuilder,
    dtype: DataType,
    prep: tuple,
) -> Tuple[Any]:
    return tuple(env.call('join_table_on', v) for v in prep)


@ufunc.resolver(SqlQueryBuilder, Bind)
def join_table_on(env: SqlQueryBuilder, item: Bind):
    prop = env.model.flatprops.get(item.name)
    if not prop or not authorized(env.context, prop, Action.SEARCH):
        raise PropertyNotFound(env.model, property=item.name)
    return env.call('join_table_on', prop)


@ufunc.resolver(SqlQueryBuilder, Bind, name='len')
def len_(env: SqlQueryBuilder, bind: Bind):
    prop = env.model.flatprops[bind.name]
    return env.call('len', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, str, name='len')
def len_(env: SqlQueryBuilder, bind: str):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    prop = env.model.flatprops[bind]
    return env.call('len', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, DataType, name='len')
def len_(env: SqlQueryBuilder, dtype: DataType):
    column = env.backend.get_column(env.table, dtype.prop)
    return sa.func.length(column)


@ufunc.resolver(SqlQueryBuilder, Expr)
def sort(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    env.sort = []
    for key in args:
        prop = env.model.properties[key.name]
        column = env.backend.get_column(env.table, prop)
        if isinstance(key, Negative):
            column = column.desc()
        else:
            column = column.asc()
        env.sort.append(column)


@ufunc.resolver(SqlQueryBuilder, int)
def limit(env: SqlQueryBuilder, n: int):
    env.limit = n


@ufunc.resolver(SqlQueryBuilder, int)
def offset(env: SqlQueryBuilder, n: int):
    env.offset = n


@ufunc.resolver(SqlQueryBuilder, Property, object, object)
def swap(env: SqlQueryBuilder, prop: Property, old: Any, new: Any) -> Any:
    return Expr('swap', old, new)


@ufunc.resolver(SqlQueryBuilder, Expr)
def file(env: SqlQueryBuilder, expr: Expr) -> Expr:
    args, kwargs = expr.resolve(env)
    return Expr(
        'file',
        name=env.call('select', kwargs['name'], nested=True),
        content=env.call('select', kwargs['content'], nested=True),
    )


class _FileSelected(TypedDict):
    name: Selected      # File name
    content: Selected   # File content


@ufunc.resolver(SqlResultBuilder, Expr)
def file(env: SqlResultBuilder, expr: Expr) -> FileData:
    """Post query file data processor

    Will be called with _FileSelected kwargs and no args.
    """
    kwargs: _FileSelected
    args, kwargs = expr.resolve(env)
    assert len(args) == 0, args
    name = env.data[kwargs['name'].item]
    content = env.data[kwargs['content'].item]
    if isinstance(content, str):
        content = content.encode('utf-8')
    if content is not None:
        content = base64.b64encode(content).decode()
    return {
        '_id': name,
        # TODO: Content probably should not be returned if not explicitly
        #       requested in select list.
        '_content': content,
    }


@ufunc.resolver(SqlQueryBuilder)
def cast(env: SqlQueryBuilder) -> Expr:
    return Expr('cast')


@ufunc.resolver(SqlResultBuilder)
def cast(env: SqlResultBuilder) -> Any:
    return env.call('cast', env.prop.dtype, env.this)


@ufunc.resolver(SqlResultBuilder, String, int)
def cast(env: SqlResultBuilder, dtype: String, value: int) -> str:
    return str(value)


@ufunc.resolver(SqlResultBuilder, String, type(None))
def cast(env: SqlResultBuilder, dtype: String, value: Optional[Any]) -> str:
    return ''



from typing import Any
from typing import Dict
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import overload

from spinta import commands
from spinta import exceptions
from spinta.components import Component
from spinta.components import Config
from spinta.components import Context
from spinta.components import EntryId
from spinta.components import Node
from spinta.manifests.components import Manifest
from spinta.utils.schema import NA
from spinta.utils.schema import resolve_schema


def get_node(
    config: Config,
    manifest: Manifest,
    # MetaData entry ID, for yaml manifests it's filename, for backend manifests
    # it's UUID, for CSV tables it's row number.
    eid: EntryId,
    data: dict = None,
    *,
    # Component group from config.components.
    group: str = 'nodes',
    # If None component name will be taken from data['type'].
    ctype: str = None,
    # If parent is None, then parent is assumed to be manifest.
    parent: Component = None,
    check: bool = True,
):
    if data is not None and not isinstance(data, dict):
        raise exceptions.InvalidManifestFile(
            parent or manifest,
            manifest=manifest.name,
            eid=eid,
            error=f"Expected dict got {type(data).__name__}.",
        )

    if ctype is None:
        if 'type' not in data:
            raise exceptions.InvalidManifestFile(
                parent or manifest,
                manifest=manifest.name,
                eid=eid,
                error=f"Required parameter 'type' is not defined.",
            )

        ctype = data['type']

    if parent is None:
        # If parent is given, that means we are loading a node whose parent is
        # not manifest, that means we can't do checks on manifest.objects.

        if ctype not in manifest.objects:
            raise exceptions.InvalidManifestFile(
                manifest=manifest.name,
                eid=eid,
                error=f"Unknown type {ctype!r}.",
            )

        if check:
            if 'name' not in data:
                raise exceptions.MissingRequiredProperty(
                    manifest,
                    schema=data['path'],
                    prop='name',
                )

            if data['name'] in manifest.objects[ctype]:
                name = data['name']
                other = manifest.objects[ctype][name].eid
                raise exceptions.InvalidManifestFile(
                    manifest=manifest.name,
                    eid=eid,
                    error=(
                        f"{ctype!r} with name {name!r} already defined in "
                        f"{other}."
                    ),
                )

    if ctype not in config.components[group]:
        from spinta.components import Property
        if group == 'types' and isinstance(parent, Property):
            if ctype:
                error = (
                    f"Unknown {ctype!r} type of {parent.place!r} property "
                    f"in {parent.model.name!r} model."
                )
            else:
                error = (
                    f"Type is not given for {parent.place!r} property "
                    f"in {parent.model.name!r} model."
                )
        else:
            error = f"Unknown component {ctype!r} in {group!r}."
        raise exceptions.InvalidManifestFile(
            parent,
            manifest=manifest.name,
            eid=eid,
            error=error,
        )

    Node_ = config.components[group][ctype]
    return Node_()


@overload
def load_node(
    context: Context,
    node: Union[Node, Component],
    data: dict,
    *,
    mixed: Literal[True] = False,
    parent: Node = None,
) -> Tuple[Node, dict]:
    ...


@overload
def load_node(
    context: Context,
    node: Union[Node, Component],
    data: dict,
    *,
    mixed: Literal[False] = False,
    parent: Node = None,
) -> Node:
    ...


def load_node(
    context: Context,
    node: Union[Node, Component],
    data: dict,
    *,
    mixed: bool = False,
    parent: Union[Node, Manifest] = None,
) -> Union[
    Node,
    Tuple[Node, Dict[str, Any]]
]:
    remainder = {}
    node_schema = resolve_schema(node, Component)
    for name in set(node_schema) | set(data):
        if name not in node_schema:
            if mixed:
                remainder[name] = data[name]
                continue
            else:
                raise exceptions.UnknownParameter(node, param=name)
        schema = node_schema[name]
        if schema.get('parent'):
            attr = schema.get('attr', name)
            assert parent is not None, node
            setattr(node, attr, parent)
            continue
        value = data.get(name, NA)
        if schema.get('inherit', False) and value is NA:
            if node.parent and hasattr(node.parent, name):
                value = getattr(node.parent, name)
            else:
                value = None
        if schema.get('required', False) and (value is NA or value is None):
            raise exceptions.MissingRequiredProperty(node, prop=name)
        if value is NA:
            if 'factory' in schema:
                value = schema['factory']()
            else:
                value = schema.get('default')
        elif schema.get('type') == 'array':
            if not isinstance(value, list) and schema.get('force'):
                value = [value]
        attr = schema.get('attr', name)
        setattr(node, attr, value)
    if mixed:
        return node, remainder
    else:
        return node


def load_model_properties(
    context: Context,
    model: Node,
    Prop: Type[Node],
    data: Optional[dict],
) -> None:
    data = data or {}

    # Add build-in properties.
    data = {
        '_op': {'type': 'string'},
        '_type': {'type': 'string'},
        '_id': {'type': 'pk', 'unique': True},
        '_revision': {'type': 'string'},
        '_txn': {'type': 'string'},
        '_cid': {'type': 'integer'},
        '_created': {'type': 'datetime'},
        '_where': {'type': 'rql'},
        **data,
    }

    model.flatprops = {}
    model.leafprops = {}
    model.properties = {}
    for name, params in data.items():
        prop = Prop()
        prop.name = name
        prop.place = name
        prop.path = model.path
        prop.model = model
        prop = commands.load(context, prop, params, model.manifest)
        model.properties[name] = prop
        model.flatprops[name] = prop



from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Object
from spinta.types.helpers import set_dtype_backend


@commands.link.register(Context, Object)
def link(context: Context, dtype: Object) -> None:
    set_dtype_backend(dtype)
    for prop in dtype.properties.values():
        commands.link(context, prop.dtype)



from spinta import commands
from spinta.components import Context, Action, DataItem, Property
from spinta.types.datatype import DataType
from spinta.exceptions import UniqueConstraint
from spinta.backends.mongo.components import Mongo


@commands.check_unique_constraint.register()
def check_unique_constraint(
    context: Context,
    data: DataItem,
    dtype: DataType,
    prop: Property,
    backend: Mongo,
    value: object,
):
    model = prop.model
    table = backend.db[model.model_type()]
    # XXX: Probably we should move this out of mongo backend and implement
    #      this on spinta.commands.write. For example, read_existing_data
    #      could try to read existing record if `_id` or any other unique
    #      field is given. Also this would fix case, when multiple
    #      properties are given as unique constraint.
    if prop.name == '_id':
        name = '__id'
    else:
        name = prop.name
    # TODO: Add support for nested properties.
    # FIXME: Exclude currently saved value.
    #        In case of an update, exclude currently saved value from
    #        uniqueness check.
    if data.action in (Action.UPDATE, Action.PATCH):
        if name == '__id' and value == data.saved['_id']:
            return

        result = table.find_one({
            '$and': [{name: value},
                     {'__id': {'$ne': data.saved['_id']}}],
        })
    else:
        result = table.find_one({name: value})
    if result is not None:
        raise UniqueConstraint(prop, value=value)



from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.backends.fs.components import FileSystem


@commands.migrate.register(Context, Manifest, FileSystem)
def migrate(context: Context, manifest: Manifest, backend: FileSystem):
    pass
